import pymongo

from ereuse_devicehub.resources.account.role import Role
from ereuse_devicehub.resources.resource import ResourceSettings
from ereuse_devicehub.resources.schema import Thing
from ereuse_devicehub.validation.validation import ALLOWED_WRITE_ROLES


class Account(Thing):
    """
    An account represents a physical person or an organization.
    """
    email = {
        'type': 'email',
        'required': True,
        'unique': True,
        'sink': 5
    }
    password = {
        'type': 'string',
        # 'required': True, todo active OR password required
        'minlength': 4,
        'sink': 4,
        'doc': 'Users can only see their own passwords.'
    }
    role = {
        'type': 'string',
        'allowed': set(Role.ROLES),
        'default': Role.BASIC,
        'doc': 'See the Roles section to get more info.',
        ALLOWED_WRITE_ROLES: Role.MANAGERS
    }
    token = {
        'type': 'string',
        'readonly': True,
    }
    name = {
        'type': 'string',
        'sink': 3,
        'description': 'The name of an account, if it is a person or an organization.'
    }
    organization = {
        'type': 'string',
        'sink': 1,
        'description': 'The name of the organization the account is in. Organizations can be inside other organizations.'
    }
    active = {
        'type': 'boolean',
        'default': True,
        'sink': -1,
        'description': 'Activate the account so you can start using it.',
        'doc': 'Inactive accounts cannot login, and they are created through regular events. `Employee` or below cannot see this parameter.'
    }
    blocked = {
        'type': 'boolean',
        'default': True,
        'sink': -2,
        'description': 'As a manager, you need to specifically accept the user by unblocking it\'s account.',
        ALLOWED_WRITE_ROLES: Role.MANAGERS
    }
    isOrganization = {
        'type': 'boolean',
        'sink': 2
    }
    databases = {  # todo set allowed for the active databases
        'type': 'databases',
        'required': True,
        ALLOWED_WRITE_ROLES: Role.MANAGERS,
        'teaser': False,
        'sink': -4,
    }
    defaultDatabase = {
        'type': 'string',  # todo If this is not set, the first databased in 'databases' it should be used
        ALLOWED_WRITE_ROLES: Role.MANAGERS,
        'teaser': False,
        'sink': -5
    }
    fingerprints = {
        'type': 'list',
        'readonly': True,
    }
    publicKey = {
        'type': 'string',
        'writeonly': True
    }


class AccountSettings(ResourceSettings):
    resource_methods = ['GET', 'POST']
    item_methods = ['PATCH', 'DELETE', 'GET']
    # the standard account entry point is defined as
    # '/accounts/<ObjectId>'. We define  an additional read-only entry
    # point accessible at '/accounts/<username>'.
    # Note that this regex is weak; it will accept more string that are not emails, which is fine; it is fast.
    additional_lookup = {
        'url': 'regex("[^@]+@[^@]+\.[^@]+")',
        'field': 'email',
    }
    # 'public_methods': ['POST'],  # Everyone can create an account, which will be blocked (not active)

    datasource = {
        'projection': {'token': 0},  # We exclude from showing tokens to everyone
        'source': 'accounts'
    }

    # We also disable endpoint caching as we don't want client apps to
    # cache account data.
    cache_control = ''
    cache_expires = 0

    # Allow 'token' to be returned with POST responses
    extra_response_fields = ResourceSettings.extra_response_fields + ['token', 'email', 'role', 'active', 'name',
                                                                      'databases', 'defaultDatabase', 'organization',
                                                                      'isOrganization']

    # Finally, let's add the schema definition for this endpoint.
    _schema = Account

    mongo_indexes = {
        'email': [('email', pymongo.DESCENDING)],
        'name': [('name', pymongo.DESCENDING)],
        'email and name': [('email', pymongo.DESCENDING), ('name', pymongo.DESCENDING)]
    }

    get_projection_blacklist = {  # whitelist has more preference than blacklist
        '*': ('password',),  # No one can see password
        Role.EMPLOYEE: ('active',)  # Regular users cannot see if someone is active or not
    }
    get_projection_whitelist = {
        'author': ('password', 'active')  # Except the own author
    }
    allowed_item_write_roles = {Role.AMATEUR}  # Amateur can write it's account
    use_default_database = True  # We have a common shared database with accounts


unregistered_user = {
    'email': Account.email,
    'name': Account.name,
    'organization': Account.organization,
    'isOrganization': Account.isOrganization
}



from ereuse_devicehub.tests.test_resources.test_events.test_device_event import TestDeviceEvent


class TestReceive(TestDeviceEvent):
    def test_create_receive_with_place(self):
        receive = self.get_fixture('receive', 'receive')
        receive['receiver'] = self.get_first('accounts')['_id']
        receive['devices'] = self.devices_id
        receive['place'] = self.place['_id']
        self.post_and_check(self.DEVICE_EVENT + '/receive', receive)



from ereuse_devicehub.resources.schema import UnitCodes, RDFS


class Benchmark(RDFS):
    BENCHMARK_HARD_DRIVE = 'BenchmarkHardDrive'
    BENCHMARK_PROCESSOR = 'BenchmarkProcessor'
    TYPES = (
        BENCHMARK_HARD_DRIVE,
        BENCHMARK_PROCESSOR
    )
    readingSpeed = {
        'type': 'float',
        'unitCode': UnitCodes.mbyte
    }
    writingSpeed = {
        'type': 'float',
        'unitCode': UnitCodes.mbyte
    }

    @classmethod
    def _clean(cls, attributes: dict, attributes_to_remove: tuple = None) -> dict:
        attributes_to_remove = tuple() if attributes_to_remove is None else attributes_to_remove
        benchmark_attributes_to_remove = ('BENCHMARK_HARD_DRIVE', 'BENCHMARK_PROCESSOR', 'TYPES')
        return super(Benchmark, cls)._clean(attributes, attributes_to_remove + benchmark_attributes_to_remove)


class BenchmarkHardDrive(Benchmark):
    pass


class BenchmarkWithScore(Benchmark):
    score = {
        'type': 'float'
    }


class BenchmarkProcessor(BenchmarkWithScore):
    pass



import pymongo
from ereuse_devicehub.resources.device.domain import DeviceDomain
from ereuse_devicehub.resources.event.device import DeviceEventDomain
from ereuse_devicehub.resources.event.device.allocate.hooks import materialize_actual_owners_add
from ereuse_devicehub.resources.event.device.deallocate.hooks import materialize_actual_owners_remove


def materialize_owners(devices_id):
    """Re-computes 'owners' for the given devices and components."""
    # First let's erase all owners
    DeviceDomain.update_many_raw({'_id': {'$in': devices_id}}, {'$set': {'owners': []}})
    # Then let's execute again the materialize hooks for Allocate/Deallocate
    query = {'$or': [{'@type': 'devices:Allocate'}, {'@type': 'devices:Deallocate'}], 'devices': {'$in': devices_id}}
    order_by = {'_created': pymongo.ASCENDING}
    for event in DeviceEventDomain.get({'$query': query, '$orderby': order_by}):
        if event['@type'] == 'devices:Allocate':
            materialize_actual_owners_add([event])
        else:
            modified = materialize_actual_owners_remove([event])
            if modified == 0:  # This Remove does nothing and should be erased
                DeviceEventDomain.delete({'_id': event['_id']})



"""
DeviceHub app
"""
import inspect
import os
import sys

import gnupg
from eve import Eve
from eve.endpoints import schema_collection_endpoint
from eve.exceptions import ConfigException
from eve.io.mongo import GridFSMediaStorage
from eve.io.mongo import MongoJSONEncoder

from ereuse_devicehub.aggregation.settings import aggregate_view
from ereuse_devicehub.data_layer import DataLayer
from ereuse_devicehub.error_handler import ErrorHandlers
from ereuse_devicehub.hooks import hooks
from ereuse_devicehub.request import RequestSignedJson
from ereuse_devicehub.resources.account.login.settings import login
from ereuse_devicehub.resources.resource import ResourceSettings
from ereuse_devicehub.resources.submitter.grd_submitter.grd_submitter import GRDSubmitter
from ereuse_devicehub.resources.submitter.submitter_caller import SubmitterCaller
from ereuse_devicehub.security.authentication import RolesAuth
from ereuse_devicehub.static import send_device_icon
from ereuse_devicehub.utils import cache
from ereuse_devicehub.validation.validation import DeviceHubValidator
from eve.render import send_response
from flask import json
from flask import request


class DeviceHub(Eve):
    def __init__(self, import_name=__package__, settings='settings.py', validator=DeviceHubValidator, data=DataLayer,
                 auth=RolesAuth, redis=None, url_converters=None, json_encoder=None, media=GridFSMediaStorage,
                 **kwargs):
        kwargs.setdefault('static_url_path', '/static')
        super().__init__(import_name, settings, validator, data, auth, redis, url_converters, json_encoder, media,
                         **kwargs)
        self.json_encoder = MongoJSONEncoder
        self.request_class = RequestSignedJson
        self.gpg = gnupg.GPG()
        self.cache = cache
        self.cache.init_app(self)
        hooks(self)  # Set up hooks. You can add more hooks by doing something similar with app "hooks(app)"
        ErrorHandlers(self)
        self.add_url_rule('/login', 'login', view_func=login, methods=['POST', 'OPTIONS'])
        self.add_url_rule('/devices/icons/<file_name>', view_func=send_device_icon)
        self.add_url_rule('/<db>/aggregations/<resource>/<method>', 'aggregation', view_func=aggregate_view)
        if self.config.get('GRD', True):
            self.grd_submitter_caller = SubmitterCaller(self, GRDSubmitter)

    def register_resource(self, resource: str, settings: ResourceSettings):
        """
            Recursively registers a resource and it's sub-resources.
        """
        if inspect.isclass(settings) and issubclass(settings, ResourceSettings):
            for sub_resource_settings in settings.sub_resources():
                self.register_resource(sub_resource_settings.resource_name(), sub_resource_settings)
            # noinspection PyCallingNonCallable
            new_settings = settings()
        else:
            new_settings = settings
        super().register_resource(resource, new_settings)

    def _add_resource_url_rules(self, resource, settings):
        """
            For the given resources set to work with different databases, it registers as many URL as defined databases
            in the following way:
            For resource 'devices' and db1, db2, ... dn databases:
            - db1/devices
            - db2/devices
            ...
            - dbn/devices

            Check the attribute 'use_default_database' in class :class:`app.resources.resource.ResourceSettings`,
            which forces the resource to just use the default database.
        """
        if settings.get('use_default_database', False):
            super(DeviceHub, self)._add_resource_url_rules(resource, settings)
        else:
            real_url_prefix = self.config['URL_PREFIX']
            for database in self.config['DATABASES']:
                if real_url_prefix:
                    self.config['URL_PREFIX'] += '/{}'.format(database)
                else:
                    self.config['URL_PREFIX'] = database
                super(DeviceHub, self)._add_resource_url_rules(resource, settings)
            self.config['URL_PREFIX'] = real_url_prefix

    def validate_roles(self, directive, candidate, resource):
        """
            The same as eve's, but letting roles to be sets, which is a more adequate type that easies some code.
        """
        roles = candidate[directive]
        if not isinstance(roles, set) and not isinstance(roles, list):
            raise ConfigException("'%s' must be set"
                                  "[%s]." % (directive, resource))

    def load_config(self):
        """
            Same as eve's, just adding our default_settings
        """

        # load defaults
        self.config.from_object('eve.default_settings')

        self.config.from_object(
            'ereuse_devicehub.default_settings')  # We just add this line todo way to avoid writing all method?

        # overwrite the defaults with custom user settings
        if isinstance(self.settings, dict):
            self.config.update(self.settings)
        else:
            if os.path.isabs(self.settings):
                pyfile = self.settings
            else:
                abspath = os.path.abspath(os.path.dirname(sys.argv[0]))
                pyfile = os.path.join(abspath, self.settings)
            try:
                self.config.from_pyfile(pyfile)
            except IOError:
                # assume envvar is going to be used exclusively
                pass
            except:
                raise

        # overwrite settings with custom environment variable
        envvar = 'EVE_SETTINGS'
        if os.environ.get(envvar):
            self.config.from_envvar(envvar)

    def _init_schema_endpoint(self):
        """Adds '_settings' field to every schema."""
        super()._init_schema_endpoint()
        self.view_functions['schema_collection'] = self._schema_endpoint

    def _schema_endpoint(self):
        """The same as 'schema_collection_endpoint', but adding '_settings' field to every schema."""
        response = schema_collection_endpoint()
        if request.method == 'OPTIONS':
            return response
        else:
            schemas = json.loads(response.data.decode())
            for resource_type, schema in schemas.items():
                settings = self.config['DOMAIN'][resource_type]
                schema['_settings'] = {
                    'url': settings['url'],
                    'use_default_database': settings['use_default_database']
                }
                try:
                    schema['_settings']['fa'] = settings['fa']
                except:
                    pass
                try:
                    schema['_settings']['short_description'] = settings['short_description']
                except:
                    pass
                try:
                    schema['_settings']['sink'] = settings['sink']
                except:
                    pass
                try:
                    schema['_settings']['icon'] = settings['icon']
                except:
                    pass
            return send_response(None, (schemas,))



import copy

from ereuse_devicehub.resources.device.schema import Device
from ereuse_devicehub.resources.device.settings import DeviceSubSettings


class Peripheral(Device):
    type = {
        'type': 'string',
        'allowed': {'Router', 'Switch', 'Printer', 'Scanner', 'MultifunctionPrinter', 'Terminal', 'HUB', 'SAI',
                    'Keyboard', 'Mouse'},
        'required': True
    }
    manufacturer = copy.copy(Device.manufacturer)
    manufacturer['required'] = True
    serialNumber = copy.copy(Device.serialNumber)
    serialNumber['required'] = True
    model = copy.copy(Device.model)
    model['required'] = True


class PeripheralSettings(DeviceSubSettings):
    _schema = Peripheral



from pprint import pprint

from ereuse_devicehub.tests.test_resources.test_events.test_device_event.test_snapshot.test_snapshot import TestSnapshot


class DummyDB:
    """
        Simple class to create a database from 0, adding a dummy user and some devices. Ready to test!

        To use it do the following in your app.py::

            app = DeviceHub()
            d = DummyDB(app)
            d.create_dummy_devices()

        This method uses the test class of TestSnapshot, and the credentials in
        :func:`ereuse_devicehub.tests.TestBase.create_dummy_user`.

    """

    def __init__(self, app):
        self.app = app
        self.test_snapshot = TestSnapshot()
        self.test_snapshot.app = self.app
        self.test_snapshot.prepare()
        self.test_snapshot.create_dummy_user()

    def create_dummy_devices(self, maximum: int = None):
        self.test_snapshot.creation = self._creation
        self.test_snapshot.test_condition(maximum)
        self.test_snapshot.post_snapshot(self.test_snapshot.get_fixture(self.test_snapshot.SNAPSHOT, 'erase_sectors'))
        self.test_snapshot.post_snapshot(self.test_snapshot.get_fixture(self.test_snapshot.SNAPSHOT,
                                                                        'erase_sectors_steps'))
        pprint('Finished basic creation of devices.')

    def _creation(self, input_snapshot, *args, **kwargs):
        self.test_snapshot.post_and_check(self.test_snapshot.DEVICE_EVENT + '/snapshot', input_snapshot)



from ereuse_devicehub.tests import TestStandard


class TestSecurity(TestStandard):
    def setUp(self, settings_file=None, url_converters=None):
        super(TestSecurity, self).setUp(settings_file, url_converters)
        self.devices_id = self.get_fixtures_computers()

    def test_access_public_devices(self):
        public_id = self.devices_id[0]
        private_id = self.devices_id[1]
        public_patch = {
            'public': True
        }
        _, status_code = self.patch(self.DEVICES + '/' + public_id, public_patch)
        self.assert200(status_code)
        # We access to the public with credentials and check that it has 'public' field and equals to True
        public, status_code = self.get(self.DEVICES, '', public_id)
        self.assert200(status_code)
        self.assertIn('public', public)
        self.assertTrue(public['public'])
        public_component, _ = self.get(self.DEVICES, '', public['components'][0])
        self.assertIn('public', public_component)
        self.assertTrue(public_component['public'])
        # We access to the public device without credentials
        _, status_code = self.get(self.DEVICES, '', public_id, False)
        self.assert200(status_code)
        # We access to the private device without credentials
        _, status_code = self.get(self.DEVICES, '', private_id, False)
        self.assert401(status_code)
        # We access to all the devices without credentials
        _, status_code = self.get(self.DEVICES, '', None, False)
        self.assert401(status_code)
        # We access to all the devices with credentials
        _, status_code = self.get(self.DEVICES, '', None, True)
        self.assert200(status_code)
        # We access to another item endpoint without credentials
        private_event = self.get_first(self.DEVICE_EVENT)
        _, status_code = self.get(self.DEVICE_EVENT, '', private_event['_id'], False)
        self.assert401(status_code)



from assertpy import assert_that

from ereuse_devicehub.resources.event.device import DeviceEventDomain
from ereuse_devicehub.tests import TestStandard


class TestDevice(TestStandard):
    def setUp(self, settings_file=None, url_converters=None):
        super().setUp(settings_file, url_converters)
        self.place = self.post_fixture(self.PLACES, self.PLACES, 'place')

    def test_materializations_events(self):
        """
        Tests materializations related to events.
        :return:
        """
        # Let's check POST
        devices = self.get_fixtures_computers()
        vaio, _ = self.get(self.DEVICES, '', devices[0])
        account_id = str(self.account['_id'])
        materialized = [
            {'@type': DeviceEventDomain.new_type('Snapshot'), 'secured': False, 'byUser': account_id, 'incidence': False},
            {'@type': DeviceEventDomain.new_type('Register'), 'secured': False, 'byUser': account_id, 'incidence': False},
        ]
        fields = {'@type', '_id', 'byUser', 'incidence', 'secured', '_updated'}
        self.assertIn('events', vaio)
        i = 0
        for event in vaio['events']:
            # Order needs to be preserved
            assert_that(materialized[i]).is_subset_of(event)
            assert_that(set(event.keys())).is_equal_to(fields)
            i += 1



from eve.utils import document_etag
from flask import current_app

from ereuse_devicehub.resources.device.exceptions import DeviceNotFound
from ereuse_devicehub.resources.device.settings import DeviceSettings
from ereuse_devicehub.resources.domain import Domain, ResourceNotFound
from ereuse_devicehub.utils import Naming


class DeviceDomain(Domain):
    resource_settings = DeviceSettings

    @classmethod
    def get_one(cls, id_or_filter: str or dict) -> dict:
        """
        :throws DeviceNotFound:
        """
        try:
            return super().get_one(id_or_filter)
        except ResourceNotFound:
            raise DeviceNotFound()

    @staticmethod
    def generate_etag(device: dict) -> str:
        domain = current_app.config['DOMAIN']
        return document_etag(device, domain[Naming.resource(device['@type'])]['etag_ignore_fields'])

    @classmethod
    def seem_equal(cls, x: dict, y: dict) -> bool:
        x_tag = x['_etag'] if '_etag' in x else cls.generate_etag(x)
        y_tag = y['_etag'] if '_etag' in y else cls.generate_etag(y)
        return x_tag == y_tag

    @classmethod
    def difference(cls, list_to_remove_devices_from, checking_list):
        """
        Computes the difference between two lists of devices.

        To compute the difference the position of the parameters is important
        :param list_to_remove_devices_from:
        :param checking_list:
        :return:
        """
        difference = []
        for x in list_to_remove_devices_from:
            found = False
            for y in checking_list:
                if cls.seem_equal(x, y):
                    found = True
            if not found:
                difference.append(x)
        return difference



from ereuse_devicehub.resources.event.device.erase_basic.settings import EraseBasic
from ereuse_devicehub.resources.event.device.settings import EventSubSettingsOneDevice


class EraseSectors(EraseBasic):
    pass


class EraseSectorsSettings(EventSubSettingsOneDevice):
    _schema = EraseSectors
    fa = 'fa-eraser'
    item_methods = ['GET']



"""
    Submits events to GRD, handling authentication and translation.
"""
import copy
import re
from urllib.parse import quote_plus


from ereuse_devicehub.resources.event.device import DeviceEventDomain
from ereuse_devicehub.resources.submitter.submitter import Submitter
from ereuse_devicehub.resources.submitter.translator import Translator
from ereuse_devicehub.security.request_auth import Auth
from ereuse_devicehub.utils import Naming
from ereuse_devicehub.validation.validation import HID_REGEX


class GRDSubmitter(Submitter):
    def __init__(self, token: str, app, **kwargs):
        config = app.config
        domain = config['GRD_DOMAIN']
        translator = GRDTranslator(config)
        account = config['GRD_ACCOUNT']
        auth = Auth(domain, account['username'], account['password'], 'api-token-auth/', 'Token')
        debug = config.get('GRD_DEBUG', False)
        super().__init__(token, app, domain, translator, auth, debug)

    def generate_url(self, original_resource, translated_resource):
        device_identifier = self.translator.hid_or_url(original_resource['device'])
        if not re.compile(HID_REGEX).match(device_identifier):  # It is not a HID, so it is an URL
            device_identifier = quote_plus(device_identifier.replace('/', '!'))  # Adapt it to GRD needs
        url = self.domain + '/api/devices/'
        event_type = translated_resource['@type']
        if event_type == DeviceEventDomain.new_type('Register'):
            url += 'register'
        else:
            url += '{}/{}'.format(device_identifier, Naming.resource(event_type))
        return url


class GRDTranslator(Translator):
    def __init__(self, config):
        generic_resource = {
            'url': (self.url('events'), '_id'),
            'date': (self.identity,),
            'byUser': (self.url('accounts'),),
            'dhDate': (self.identity, '_created'),
            'errors': (self.identity, '_errors'),
            'secured': (self.identity,),
            'incidence': (self.identity,),
            'device': (self.hid_or_url,),
            'geo': (self.identity,),
            '@type': (self.identity,)
        }
        prefix = DeviceEventDomain.new_type
        translation_dict = {
            prefix('Register'): {
                'offline': (self.identity,),
                'device': (self.device,),
                'components': (self.for_all(self.device),)
            },
            prefix('Deallocate'): {
                'from': (self.url('accounts'),)
            },
            prefix('Allocate'): {
                'to': (self.url('accounts'),)
            },
            prefix('UsageProof'): {
                'softwareVersion': (self.identity,)
            },
            prefix('Receive'): {
                'receiver': (self.url('accounts'),),
                'type': (self.identity,),
                'place': (self.url('places'),)
            },
            prefix('Add'): {
                'components': (self.for_all(self.hid_or_url),)
            },
            prefix('Remove'): {
                'components': (self.for_all(self.hid_or_url),)
            },
            prefix('Locate'): {
                'place': (self.url('places'),)
            },
            prefix('Recycle'): dict(),
            prefix('Migrate'): dict()
        }
        super().__init__(config, generic_resource, translation_dict)

    def translate(self, database: str, event: dict) -> list:
        self.database = database
        translated = []
        if 'devices' in event and event['@type'] != DeviceEventDomain.new_type('Register'):
            e = copy.deepcopy(event)
            del e['devices']
            for device in event['devices']:
                e['device'] = device
                translated.append((self._translate(e), copy.deepcopy(e)))
        else:
            translated.append((self._translate(event), event))
        return translated

    def device(self, device: dict) -> dict:
        grd_device = {
            '@type': device['@type'],
            'url': self._get_resource_url(device['_id'], 'devices'),
        }
        if 'pid' in device:
            grd_device['pid'] = device['pid']
        if 'hid' in device:
            grd_device['hid'] = device['hid']
        return super().device(grd_device)




from flask import g

from ereuse_devicehub.resources.device.component.domain import ComponentDomain
from ereuse_devicehub.resources.device.domain import DeviceDomain
from ereuse_devicehub.resources.device.exceptions import DeviceNotFound, NoDevicesToProcess
from ereuse_devicehub.resources.event.device import DeviceEventDomain
from ereuse_devicehub.rest import execute_post
from ereuse_devicehub.utils import Naming
from .event_processor import EventProcessor


class Snapshot:
    def __init__(self, device: dict, components: list):
        self.events = EventProcessor()
        self.device = device
        self.components = components
        self.unsecured = []
        self.test_hard_drives = g.snapshot_test_hard_drives = []
        self.erasures = g.snapshot_basic_erasures = []

    def execute(self):
        event_log = []
        self.get_tests_and_erasures(self.components)
        self.register(event_log)
        for component in self.components:
            self.get_add_remove(component, self.device)
        self.exec_hard_drive_events(event_log, self.erasures + self.test_hard_drives)
        self._remove_nonexistent_components()
        return event_log + self.events.process()

    def get_add_remove(self, device: dict, new_parent: dict):
        """
        Get the changes (events) that will need to be triggered for the given device.
        Changes will we saved in the same device in the reserved key '_change'.
        The function doesn't execute any event per se or validate that the user can do it.

        :param device: The device must have an hid
        :param new_parent:
        """
        if 'new' not in device:
            try:
                old_parent = ComponentDomain.get_parent(device['_id'])
            except DeviceNotFound:  # The component exists but had no parent device, until now
                self.events.append_add(device, new_parent)
            else:
                if not DeviceDomain.seem_equal(old_parent, new_parent):
                    self.events.append_remove(device, old_parent)
                    self.events.append_add(device, new_parent)
        else:
            del device['new']

    def _remove_nonexistent_components(self):
        """
        This function needs to be executed at the end, after all components are in the DB and have _id

        In this stage, we have already added the components to the parent device. In the materialized components field
        of device we still have the old components. We need to remove those that are not present in this new snapshot.
        """
        if 'components' in self.device:
            full_old_components = [ComponentDomain.get_one(component) for component in self.device['components']]
            for component_to_remove in ComponentDomain.difference(full_old_components, self.components):
                self.events.append_remove(component_to_remove, self.device)

    def _append_unsecured(self, device: dict, resource_type: str):
        self.unsecured.append({'@type': device['@type'], 'type': resource_type, '_id': device['_id']})

    def register(self, event_log: list):
        try:
            register = {
                '@type': DeviceEventDomain.new_type('Register'),
                'device': self.device,
                'components': self.components
            }
            event_log.append(execute_post(Naming.resource(register['@type']), register))
        except NoDevicesToProcess:  # As it is a custom exception we throw, it keeps being an exception through post_internal
            pass
        for device in [self.device] + self.components:
            if 'hid' not in device and 'pid' not in device:
                self._append_unsecured(device, 'model')
            elif 'pid' in device:
                self._append_unsecured(device, 'pid')

    def get_tests_and_erasures(self, components):
        i = 0
        for component in components:
            if 'test' in component:
                self.test_hard_drives.append((i, component['test']))
                del component['test']
            if 'erasure' in component:
                self.erasures.append((i, component['erasure']))
                del component['erasure']
            i += 1

    def exec_hard_drive_events(self, event_log, events):
        for i, event in events:
            event['device'] = self.components[i]['_id']
            event_log.append(execute_post(Naming.resource(event['@type']), event))
            event.update(event_log[-1])



# Load the schema and resource classes
from .add import settings
from .allocate import settings
from .deallocate import settings
from .erase_basic import settings
from .erase_sectors import settings
from .generic_settings import *
from .locate import settings
from .receive import settings
from .register import settings
from .remove import settings
from .snapshot import settings
from .test_hard_drive import settings




"""
We mimic https://schema.org/unitCode, in concrete
the UN/CEFACT Common Code.
"""
import copy

from ereuse_devicehub.resources.resource import Resource
from ereuse_devicehub.utils import Naming, NestedLookup, ClassProperty


class UnitCodes:
    mbyte = '4L'
    mbps = 'E20'
    mhz = 'MHZ'
    gbyte = 'E34'
    ghz = 'A86'
    bit = 'A99'
    kgm = 'KGM'
    m = 'MTR'

    @classmethod
    def humanize(cls, code_to_search):
        for human_name, code in vars(cls).items():
            if code == code_to_search:
                return human_name


class RDFS(Resource):
    label = {
        'type': 'string',
        'sink': 5,
        'description': 'A short, descriptive title'
    }
    _type = {
        'type': 'string',
        'required': True,
        'teaser': False
    }
    _settings = {
        'prefix': None, """JSON-LD prefix to use in type. Override it to use prefix for the class."""
        'abstract': True,
        'attributes_to_remove': ('_settings', '_import_schemas', '_types', 'resource_types',
                                 'type_name', 'types', 'resource_name')
    }
    _import_schemas = True

    @classmethod
    def actual_fields(cls):
        """
        Returns the fields of only this class.
        """
        fields = super(RDFS, cls).actual_fields()
        if cls._import_schemas:
            references = []
            NestedLookup(fields, references, NestedLookup.is_sub_type_factory(RDFS))
            for document, ref_key in references:
                document[ref_key] = document[ref_key]()
        return fields

    @staticmethod
    def __new__(cls, import_schemas=True) -> dict:
        """
            Returns all the attributes of all the heriarchy: sub-classess and super-classes.
        """
        RDFS._import_schemas = import_schemas  # todo not thread safe
        full_dict = cls.superclasses_fields(2)  # We ignore Object and Resource
        allowed = full_dict['@type']['allowed']
        full_dict.update(cls.subclasses_fields())
        full_dict['@type']['allowed'] |= allowed
        return full_dict

    @classmethod
    def subclasses_fields(cls):
        """
            Returns the attributes of the subclasses, except the actual one.

            This method automatically gathers and computes @type and type attributes. We remember that, for the @type
            and type of a superclass of n subclasses, they will both have all the 'allowed' attributes
            of the n subclasses.
        """
        fields = {}
        type_allowed = set()
        subtype_allowed = set()
        for subclass in cls.subclasses():
            subclass_fields = subclass.actual_fields()
            type_allowed |= subclass_fields.get('@type', {}).get('allowed', set())
            subtype_allowed |= set(subclass_fields.get('type', {}).get('allowed', set()))
            fields.update(subclass_fields)
        if '@type' in fields:
            fields['@type']['allowed'] |= type_allowed
        if 'type' in fields:
            fields['type']['allowed'] |= subtype_allowed
        return fields

    @classmethod
    def _clean(cls, attributes: dict, attributes_to_remove: tuple = None) -> dict:
        """
            Extends :func:`Resource._clean` by setting @type accordingly and adding the 'allowed' property.
        """
        attributes_to_remove = tuple() if attributes_to_remove is None else attributes_to_remove
        fields = super()._clean(attributes, attributes_to_remove + cls._settings['attributes_to_remove'])
        if '_type' in fields:
            fields['@type'] = fields['_type']
            del fields['_type']
            if 'allowed' not in fields['@type']:
                fields['@type']['allowed'] = {cls.type_name}
        else:
            fields['@type'] = copy.deepcopy(RDFS._type)
            fields['@type']['allowed'] = {cls.type_name}
        return fields

    # noinspection PyNestedDecorators
    @ClassProperty
    @classmethod
    def resource_name(cls):
        return Naming.resource(cls.type_name)

    # noinspection PyNestedDecorators
    @ClassProperty
    @classmethod
    def type_name(cls):
        return Naming.new_type(cls.__name__, cls._settings['prefix'])

    """The following methods are not used to build the schema"""

    # noinspection PyNestedDecorators
    @ClassProperty
    @classmethod
    def types(cls):
        """
            Obtains the resource type (e.g. Accept) of the actual class and its subclasses.

            To use this method override before the attribute _settings['prefix'] and
            optionally override _settings['force_prefix'] which defaults at True.

            Read-only.
        """
        return {_class.type_name for _class in cls.subclasses() + [cls]}

    # noinspection PyNestedDecorators
    @ClassProperty
    @classmethod
    def resource_types(cls):
        return {Naming.resource(type_name) for type_name in cls.types}


class Thing(RDFS):
    url = {
        'type': 'url',
        'teaser': False,
        'doc': 'The url of the resource. If passed in, the value it is moved to sameAs.',
        'move': 'sameAs'
    }
    sameAs = {
        'type': 'url',
        'teaser': False,
        'readonly': True,
        'unique': True,
    }
    description = {
        'type': 'string',
        'maxlength': 500,
        'sink': -4,
        'description': 'Full long description'
    }




from flask import current_app as app


class RePostUsingLabel:
    is_active = False
    collection = []
    db = app.data.driver.db

    @classmethod
    def activate(cls):
        cls.is_active = True

    @classmethod
    def deactivate(cls):
        cls.is_active = False

    @classmethod
    def get_devices_and_labels(cls):
        snapshots = cls.db.events.find({'@type': 'snapshot'})
        for snapshot in snapshots:
            cls.collection.append((cls.db.devices.find_one({snapshot['device']}), snapshot['label']))

    @classmethod
    def drop(cls):
        cls.db.events.drop()
        cls.db.devices.drop()



from ereuse_devicehub.resources.device.component.settings import Component, ComponentSubSettings


class OpticalDrive(Component):
    pass


class OpticalDriveSettings(ComponentSubSettings):
    _schema = OpticalDrive



from ereuse_devicehub.exceptions import RequestAnother
from ereuse_devicehub.resources.device.domain import DeviceDomain
from ereuse_devicehub.resources.device.schema import Device
from ereuse_devicehub.resources.event.device import DeviceEventDomain
from ereuse_devicehub.resources.event.device.settings import DeviceEvent
from ereuse_devicehub.rest import execute_delete
from ereuse_devicehub.utils import Naming
from eve.utils import document_etag
from flask import current_app as app


def generate_etag(resource: str, items: list):
    if resource in Device.resource_types:
        for item in items:
            item['_etag'] = document_etag(item,
                                          app.config['DOMAIN'][Naming.resource(item['@type'])]['etag_ignore_fields'])


def post_benchmark(resource: str, devices: list):
    """
    Sets the result of one benchmark into 'benchmarks' array.

    Use this method only in POST.
    :param resource:
    :param devices:
    """
    if resource in Device.resource_types:
        for device in devices:
            if 'benchmark' in device:
                device['benchmarks'] = [device['benchmark']]
                del device['benchmark']


def autoincrement(resource: str, devices: list):
    if resource in Device.resource_types:
        for device in devices:
            device['_id'] = str(
                get_next_sequence())  # string makes this compatible with other systems that use custom id


def get_next_sequence():
    return app.data.driver.db.device_sequence.find_and_modify(
        query={'_id': 1},
        update={'$inc': {'seq': 1}},
        new=True,
        upsert=True
    ).get('seq')


def materialize_public_in_components(resource: str, devices: list):
    """
    Materializes the 'public' field of the passed in devices in their components.
    :param resource:
    :param devices:
    :return:
    """
    if resource in Device.resource_types:
        for device in devices:
            if device.get('components'):  # If empty do not try to execute
                DeviceDomain.update_raw(device['components'], {'$set': {'public': device.get('public', False)}})


def materialize_public_in_components_update(resource: str, device: dict, original: dict):
    """
    The same as :func materialize_public_in_components: but thought to be used with PATCH methods.
    :param resource:
    :param device:
    :param original:
    """
    # PATCH doesn't need to include components if we are not changing them
    # We have already saved the device to the DB so we can securely modify the device dictionary
    if original['@type'] in Device.types:
        if 'components' not in device:
            device['components'] = original['components']
        materialize_public_in_components(Naming.resource(original['@type']), [device])


class MaterializeEvents:
    """
        Materializes some fields of the events in the affected device, benefiting searches. To keep minimum space,
        only selected fields are materialized (which you can check in the following tuple)
    """
    fields = {
        '_id', '@type', 'label', 'date', 'incidence', 'secured', 'comment', 'success', 'error', 'type', 'receiver',
        'receiverOrganization', 'to', 'toOrganization', 'secured', 'byUser', 'geo', '_updated'
    }

    @classmethod
    def materialize_events(cls, resource: str, events: list):
        if resource in DeviceEvent.resource_types:
            for event in events:
                trimmed_event = {field_name: event[field_name] for field_name in cls.fields if field_name in event}
                query = {'$push': {'events': {'$each': [trimmed_event], '$position': 0}}}
                devices = [event['device']] if 'device' in event else event['devices']
                if 'parent' in event:  # Let's materialize the events (test, erasure...) of the component to the parent
                    devices.append(event['parent'])
                DeviceDomain.update_raw(devices, query)
                DeviceDomain.update_raw(event.get('components', []), query)

    @classmethod
    def dematerialize_event(cls, resource: str, event: dict):
        if event.get('@type') in DeviceEvent.types:
            device = [event['device']] if 'device' in event else []
            parent = [event['parent']] if 'parent' in device else []
            for device_id in event.get('devices', []) + event.get('components', []) + device + parent:
                DeviceDomain.update_raw(device_id, {'$pull': {'events': {'_id': event['_id']}}})


def redirect_to_first_snapshot(resource, request, lookup):
    """
    DELETE /device should be an internal method, but it is open to redirect to the door that is going to effictevely
    delete the device; this is the first Snapshot that made the Register that made the device.
    """
    if resource in Device.resource_types:
        snapshot_id = str(DeviceEventDomain.get_first_snapshot(lookup['_id'])['_id'])
        execute_delete(Naming.resource('devices:Snapshot'), snapshot_id)
        raise RequestAnother('', 204)



from ereuse_devicehub.resources.device.benchmark_settings import BenchmarkProcessor
from ereuse_devicehub.resources.device.component.settings import Component, ComponentSubSettings
from ereuse_devicehub.resources.schema import UnitCodes


class Processor(Component):
    numberOfCores = {
        'type': 'integer',
        'min': 1,
        'sink': 1
    }
    speed = {
        'type': 'float',
        'unitCode': UnitCodes.ghz,
        'sink': 1
    }
    address = {
        'type': 'integer',
        'unitCode': UnitCodes.bit,
        'allowed': {8, 16, 32, 64, 128, 256},
        'sink': -1
    }
    benchmark = {
        'type': 'dict',
        'schema': BenchmarkProcessor,
        'writeonly': True,
    }
    benchmarks = {
        'type': 'list',
        'schema': {
            'type': 'dict',
            'schema': BenchmarkProcessor
        },
        'readonly': True
    }


class ProcessorSettings(ComponentSubSettings):
    _schema = Processor



import linecache
import sys

import inflection as inflection
from flask.ext.cache import Cache

cache = Cache(config={'CACHE_TYPE': 'simple'})


class Naming:
    TYPE_PREFIX = ':'
    RESOURCE_PREFIX = '_'

    """
        In DeviceHub there are many ways to name the same resource (yay!), this is because of all the different
        types of schemas we work in. But no worries, we offer easy ways to change between naming conventions.

        - TypeCase (or resource-type) is the one represented with '@type' and follow PascalCase and always singular. This is the standard preferred one.
        - resource-case is the eve naming, using the standard URI conventions. This one is tricky, as although the types
        are represented in singular, the URI convention is to be plural (Event vs events), however just few of them
        follow this rule (Snapshot [type] to snapshot [resource]). You can set which ones you want to change their number.
        - python_case is the one used by python for its folders and modules. It is underscored and always singular.

    """

    @staticmethod
    def resource(string: str):
        """
            :param string: String can be type, resource or python case
        """
        try:
            prefix, resulting_type = Naming.pop_prefix(string)
            prefix += Naming.RESOURCE_PREFIX
        except IndexError:
            prefix = ''
            resulting_type = string
        resulting_type = inflection.dasherize(inflection.underscore(resulting_type))
        return prefix + (inflection.pluralize(resulting_type) if Naming._pluralize(resulting_type) else resulting_type)

    @staticmethod
    def python(string: str):
        """
            :param string: String can be type, resource or python case
        """
        return inflection.underscore(inflection.singularize(string) if Naming._pluralize(string) else string)

    @staticmethod
    def _pluralize(string: str):
        from ereuse_devicehub.default_settings import RESOURCES_CHANGING_NUMBER
        value = inflection.dasherize(inflection.underscore(string))
        return value in RESOURCES_CHANGING_NUMBER or inflection.singularize(value) in RESOURCES_CHANGING_NUMBER

    @staticmethod
    def type(string: str):
        try:
            prefix, resulting_type = Naming.pop_prefix(string)
            prefix += Naming.TYPE_PREFIX
        except IndexError:
            prefix = ''
            resulting_type = string
        resulting_type = inflection.singularize(resulting_type) if Naming._pluralize(resulting_type) else resulting_type
        return prefix + inflection.camelize(resulting_type)

    @staticmethod
    def url_word(word: str):
        """
            Normalizes a full word to be inserted to an url. If the word has spaces, etc, is used '_' and not '-'
        """
        return inflection.parameterize(word, '_')

    @staticmethod
    def pop_prefix(string: str):
        """Erases the prefix and returns it.
        :throws IndexError: There is no prefix.
        :return A set with two elements: 1- the prefix, 2- the type without it.
        """
        result = string.split(Naming.TYPE_PREFIX)
        if len(result) == 1:
            result = string.split(Naming.RESOURCE_PREFIX)
            if len(result) == 1:
                raise IndexError()
        return result

    @staticmethod
    def new_type(type_name: str, prefix: str or None = None) -> str:
        """
            Creates a resource type with optionally a prefix.

            Using the rules of JSON-LD, we use prefixes to disambiguate between different types with the same name:
            one can Accept a device or a project. In eReuse.org there are different events with the same names, in
            linked-data terms they have different URI. In eReuse.org, we solve this with the following:

                "@type": "devices:Accept" // the URI for these events is 'devices/events/accept'
                "@type": "projects:Accept"  // the URI for these events is 'projects/events/accept
                ...

            Type is only used in events, when there are ambiguities. The rest of

                "@type": "devices:Accept"
                "@type": "Accept"

            But these not:

                "@type": "projects:Accept"  // it is an event from a project
                "@type": "Accept"  // it is an event from a device
        """
        if Naming.TYPE_PREFIX in type_name:
            raise TypeError('Cannot create new type: type {} is already prefixed.'.format(type_name))
        prefix = (prefix + Naming.TYPE_PREFIX) if prefix is not None else ''
        return prefix + type_name


class NestedLookup:
    @staticmethod
    def __new__(cls, document, references, operation):
        """Lookup a key in a nested document, return a list of values
        From https://github.com/russellballestrini/nested-lookup/ but in python 3
        """
        return list(NestedLookup._nested_lookup(document, references, operation))

    @staticmethod
    def key_equality_factory(key_to_find):
        def key_equality(key, value):
            return key == key_to_find

        return key_equality

    @staticmethod
    def is_sub_type_factory(type):
        def _is_sub_type(key, value):
            return is_sub_type(value, type)

        return _is_sub_type

    @staticmethod
    def _nested_lookup(document, references, operation):
        """Lookup a key in a nested document, yield a value"""
        if isinstance(document, list):
            for d in document:
                for result in NestedLookup._nested_lookup(d, references, operation):
                    yield result

        if isinstance(document, dict):
            for k, v in document.items():
                if operation(k, v):
                    references.append((document, k))
                    yield v
                elif isinstance(v, dict):
                    for result in NestedLookup._nested_lookup(v, references, operation):
                        yield result
                elif isinstance(v, list):
                    for d in v:
                        for result in NestedLookup._nested_lookup(d, references, operation):
                            yield result


def is_sub_type(value, resource_type):
    try:
        return issubclass(value, resource_type)
    except TypeError:
        return issubclass(value.__class__, resource_type)


def get_last_exception_info():
    exc_type, exc_obj, tb = sys.exc_info()
    f = tb.tb_frame
    lineno = tb.tb_lineno
    filename = f.f_code.co_filename
    linecache.checkcache(filename)
    line = linecache.getline(filename, lineno, f.f_globals)
    return 'EXCEPTION IN ({}, LINE {} "{}"): {}'.format(filename, lineno, line.strip(), exc_obj)


def coerce_type(fields: dict):
    """
    Similar to a Cerberus' coercion, adds a prefix to all @types accordingly.
    :param fields: the resource (ex-JSON document) to coerce. The variable is updated.
    """
    # todo this method should be general: obtaining which resources need to be prefixed from their schema
    from ereuse_devicehub.resources.event.device import DeviceEventDomain
    references = []
    NestedLookup(fields, references, NestedLookup.key_equality_factory('@type'))
    for document, ref_key in references:
        document[ref_key] = DeviceEventDomain.add_prefix(document[ref_key])


def get_header_link(resource_type: str) -> ():
    return 'Link', '<http://www.ereuse.org/onthology/' + resource_type + '.jsonld>; rel="http://www.w3.org/ns/json-ld#context"; type="application/ld+json"'


class ClassProperty(property):
    """Defines *getting* properties from classes in python, from http://stackoverflow.com/a/29994957/2710757"""

    def __get__(self, cls, owner):
        return self.fget.__get__(None, owner)()



from bson.objectid import ObjectId

from ereuse_devicehub.resources.domain import Domain, ResourceNotFound
from ereuse_devicehub.resources.event.device.settings import EventSettings


class EventDomain(Domain):
    resource_settings = EventSettings

    @classmethod
    def get_one(cls, id_or_filter: dict or ObjectId or str, **kwargs):
        try:
            return super().get_one(id_or_filter)
        except ResourceNotFound:
            raise EventNotFound()


class EventNotFound(ResourceNotFound):
    pass



import copy

from ereuse_devicehub.resources.account.role import Role
from ereuse_devicehub.resources.account.settings import unregistered_user
from ereuse_devicehub.resources.condition import condition
from ereuse_devicehub.resources.device.component.settings import Component
from ereuse_devicehub.resources.event.device.settings import place, EventWithOneDevice, EventSubSettingsOneDevice


class Snapshot(EventWithOneDevice):
    offline = {
        'type': 'boolean'
    }
    automatic = {
        'type': 'boolean'
    }
    version = {
        'type': 'version',
        'teaser': False
    }
    snapshotSoftware = {
        'type': 'string',
        'allowed': ['DDI', 'Scan', 'DeviceHubClient'],
        'default': 'DDI'
    }
    events = {
        'type': 'list',  # Snapshot generates this automatically
        'schema': {
            'type': 'objectid',
            'data_relation': {
                'resource': 'events',
                'embeddable': True,
                'field': '_id'
            }
        },
        'readonly': True
    }
    request = {
        'type': 'string',  # The request sent, saved in case of debugging
        'readonly': True
    }
    unsecured = {
        'type': 'list',  # When we match an existing non-hid device, we state it here
        'schema': {
            'type': 'dict',
            'schema': {
                '_id': {
                    'type': 'string',
                    'data_relation': {
                        'resource': 'devices',
                        'field': '_id',
                        'embeddable': True
                    }
                },
                '@type': {
                    'type': 'string'
                },
                'type': {
                    'type': 'string',
                    'allowed': {'model', 'pid'}
                }
            }
        },
        'default': [],
        'readonly': True
    }
    device = {
        'type': 'dict',  # eve doesn't care about the type when GET values
        'required': True,
        'data_relation': {
            'resource': 'devices',
            'field': '_id',
            'embeddable': True
        },
        'sink': 4
    }
    components = {
        'type': 'list',
        'schema': {
            'type': 'dict',
            'schema': Component,
            'data_relation': {
                'resource': 'devices',
                'field': '_id',
                'embeddable': True
            }
        },
        'default': [],
        'teaser': False
    }
    debug = {
        'type': 'dict',
        'teaser': False
    }
    place = copy.deepcopy(place)
    place['description'] = 'Place the devices to an existing location.'
    place['label'] = 'Place where the devices are saved'
    software = {
        'type': 'dict',
        'schema': {
            'productKey': {
                'type': 'string'
            }
        },
        'sink': -1
    }
    condition = {
        'type': 'dict',
        'schema': condition,
        'sink': 1,
        'teaser': True
    }
    _from = {
        'type': ['objectid', 'dict', 'string'],  # We should not add string but it does not work otherwise...
        'data_relation': {
            'resource': 'accounts',
            'field': '_id',
            'embeddable': True,
        },
        'get_from_data_relation_or_create': 'email',
        'schema': unregistered_user,
        'sink': 2,
        'label': 'E-mail of the giver',
        'description': 'The e-mail of the person or organization that gave the devices. You cannot change this later.'
    }

    @classmethod
    def _clean(cls, attributes: dict, attributes_to_remove: tuple = None) -> dict:
        full_dict = super(Snapshot, cls)._clean(attributes, attributes_to_remove)
        full_dict['from'] = full_dict['_from']
        del full_dict['_from']
        return full_dict


class SnapshotSettings(EventSubSettingsOneDevice):
    _schema = Snapshot
    get_projection_blacklist = {Role.ADMIN: ('request',)}
    extra_response_fields = EventSubSettingsOneDevice.extra_response_fields + ['events']
    fa = 'fa-camera'
    short_description = "A fast picture of the state and key information of the computer and it's devices."



from ereuse_devicehub.resources.device.component.settings import Component, ComponentSubSettings


class Motherboard(Component):
    totalSlots = {
        'type': 'integer',
        'teaser': False
    }
    usedSlots = {
        'type': 'integer',
        'teaser': False
    }
    connectors = {
        'type': 'dict',
        'schema': {
            'usb': {
                'type': 'natural'
            },
            'firewire': {
                'type': 'natural'
            },
            'serial': {
                'type': 'natural'
            },
            'pcmcia': {
                'type': 'natural'
            }
        }
    }
    maxAcceptedMemory = {
        'type': 'integer',  # Maximum accepted memory that the motherboard can hold
        'teaser': False
    }


class MotherboardSettings(ComponentSubSettings):
    _schema = Motherboard
    etag_ignore_fields = ComponentSubSettings.etag_ignore_fields + ['usedSlots']



from ereuse_devicehub.exceptions import StandardError, SchemaError
from ereuse_devicehub.resources.domain import ResourceNotFound


class HidError(StandardError):
    status_code = 422
    title = 'Cannot compute hid'


class DeviceAlreadyExists(SchemaError):
    status_code = 422

    def __init__(self, field, _id):
        self.field = field
        self.message = self._id = _id


class DeviceNotFound(ResourceNotFound):
    pass


class NoDevicesToProcess(StandardError):
    status_code = 400



from ereuse_devicehub.resources.event.device.settings import components, EventWithOneDevice, EventSubSettingsOneDevice


class Remove(EventWithOneDevice):
    components = components


class RemoveSettings(EventSubSettingsOneDevice):
    _schema = Remove
    fa = 'fa-minus-square-o'
    short_description = 'Components have been removed from a device'
    item_methods = ['GET']



step = {
    '@type': {
        'type': 'string',
        'allowed': {'Zeros', 'Random'},
        'required': True
    },
    'success': {
        'type': 'boolean',
        'required': True
    },
    'startingTime': {
        'type': 'datetime'
    },
    'endingTime': {
        'type': 'datetime'
    },
    'secureRandomSteps': {
        'type': 'boolean'
    },
    'cleanWithZeros': {
        'type': 'boolean'
    }
}



from flask import Response, redirect as flask_redirect, request
from flask.json import jsonify

from ereuse_devicehub.exceptions import BasicError, RedirectToClient, RequestAnother
from ereuse_devicehub.flask_decorators import crossdomain
from ereuse_devicehub.utils import get_header_link


class ErrorHandlers:
    """
        Handles the way errors interact with the user. For example, how they are shown.
    """

    def __init__(self, app):
        @app.errorhandler(BasicError)
        @crossdomain(origin='*', headers=['Content-Type', 'Authorization'])
        def handle_standard_error(error: BasicError) -> Response:
            """
                Adapts the visualization of an standard error so it complies with the schema, Hydra.
            """
            response = jsonify(error.to_dict())
            header_name, header_value = get_header_link(type(error).__name__)
            response.headers[header_name] = header_value
            response.status_code = error.status_code
            return response

        @app.errorhandler(RedirectToClient)
        @crossdomain(origin='*', headers=['Content-Type', 'Authorization'])
        def redirect(error: RedirectToClient) -> Response:
            """
                For Redirect exceptions, performs a redirection to the client.
            """
            return flask_redirect(app.config['CLIENT'] + request.full_path)



import pymongo
from ereuse_devicehub.resources.device.component.domain import ComponentDomain
from ereuse_devicehub.resources.device.schema import Device
from ereuse_devicehub.resources.event.device import DeviceEventDomain
from ereuse_devicehub.resources.event.device.settings import Event, DeviceEvent
from ereuse_devicehub.resources.place.domain import PlaceDomain, NoPlaceForGivenCoordinates, \
    CoordinatesAndPlaceDoNotMatch
from ereuse_devicehub.rest import execute_patch, execute_delete
from ereuse_devicehub.utils import Naming
from flask import current_app


def get_place(resource_name: str, events: list):
    """

    :param resource_name:
    :param events:
    :return:
    """
    if resource_name in Event.resource_types:
        for event in events:
            if 'geo' in event:
                try:
                    place = PlaceDomain.get_with_coordinates(event['geo']['coordinates'])
                except (KeyError, NoPlaceForGivenCoordinates) as e:
                    # Receive and Locate are forced to have a place for their coordinates
                    if event['@type'] in (DeviceEventDomain.new_type(x) for x in ('Receive', 'Locate')):
                        raise e
                else:
                    if 'place' in event:
                        if event['place']['_id'] != str(place['_id']):  # geo 1 place 1
                            raise CoordinatesAndPlaceDoNotMatch()
                    else:
                        event['place'] = place['_id']  # geo 1 place found in DB


def materialize_components(resource_name: str, events: list):
    """
    Materializes the field 'components' of selected events (not all of them) with the union of all the affected
    components, when the event is performed to computers
    :param resource_name:
    :param events:
    :return:
    """
    if resource_name in Event.resource_types:
        for event in events:
            sub_schema = current_app.config['DOMAIN'][resource_name]['schema']
            if 'components' in sub_schema and sub_schema['components'].get('readonly', False):
                event['components'] = list(ComponentDomain.get_components_in_set(event['devices']))


def materialize_parent(resource_name: str, events: list):
    """
    Materializes the field 'parent' of events that only affect components (such as TestHardDrive or EraseBasic)
    :param resource_name:
    :param events:
    :return:
    """
    if resource_name in Event.resource_types:
        for event in events:
            sub_schema = current_app.config['DOMAIN'][resource_name]['schema']
            if 'parent' in sub_schema:
                event['parent'] = ComponentDomain.get_parent(event['device'])['_id']


def set_place(resource_name: str, events: list):
    """
    Sets the place of the devices. This method must execute after 'get_place' of this module.

    The event performs PATCH of place, so the effect is like setting the devices to the place.
    :param resource_name:
    :param events:
    :return:
    """
    if resource_name in Event.resource_types:
        for event in events:
            if 'place' in event:
                place = PlaceDomain.get_one(event['place'])
                device = [event['device']] if 'device' in event else []
                execute_patch('places', {'devices': list(set(place['devices'] + event.get('devices', []) + device))}, event['place'])


def unset_place(resource_name: str, event: dict):
    if resource_name in Event.resource_types:
        if 'place' in event:
            place = PlaceDomain.get_one(event['place'])
            device = [event['device']] if 'device' in event else []
            devices = event.get('devices', []) + device
            execute_patch('places', {'devices': list(set(place['devices']) - set(devices))}, event['place'])


def delete_events_in_device(resource_name: str, device: dict):
    """
    Deletes the references of the given device in all the events, and deletes the full event if it references only to
    the device.
    """
    if resource_name in Device.resource_types:
        _id = device['_id']
        qin = {'$in': [_id]}
        query = {'$or': [{'device': _id}, {'devices': qin}, {'components': qin}], '@type': {'$ne': 'devices:Register'}}
        sort = {'_created': pymongo.ASCENDING}  # Order is important to find the first Snapshot (see below)
        first_snapshot_found = False
        for event in DeviceEventDomain.get({'$query': query, '$orderby': sort}):
            if not first_snapshot_found and event['@type'] == 'devices:Snapshot':
                # We cannot delete the Snapshot that created the device, because there is a change to create
                # an infinite loop: Snapshot that created device -> Register -> DEL /device -> Snapshot that created...
                first_snapshot_found = True
            else:
                event_resource = Naming.resource(event['@type'])
                if event.get('device', None) == _id:  # Am I the 'device' of the event?
                    execute_delete(event_resource, event['_id'])
                elif [_id] == event.get('devices', []):  # Is there no more 'devices' in the event, apart from me?
                    execute_delete(event_resource, event['_id'])
                # All events that do not use 'components' for materialization (aka Add/Remove) should be erased
                # if there are no more components
                elif event['@type'] in ('devices:Add', 'devices:Remove') and event['components'] == [_id]:
                    execute_delete(event_resource, event['_id'])
                else:  # Keep the event; just delete my reference
                    DeviceEventDomain.update_raw(event['_id'], {'$pull': {'devices': qin, 'components': qin}})


def remove_from_other_events(resource_name: str, event: dict):
    """Removes the event from other events (Snapshot's 'event' per example)"""
    if resource_name in DeviceEvent.resource_types:
        update = {'$pull': {'events': {'$in': [event['_id']]}}}
        DeviceEventDomain.update_many_raw({}, update)



import os

from flask import send_from_directory

from ereuse_devicehub.flask_decorators import cache


@cache(expires=604800)
def send_device_icon(file_name):
    package_dir = os.path.abspath(os.path.dirname(__file__))
    return send_from_directory(os.path.join(package_dir, 'resources', 'device', 'icons'), file_name)



import copy

from ereuse_devicehub.resources.account.settings import unregistered_user
from ereuse_devicehub.resources.event.device.settings import place, components, EventWithDevices, \
    EventSubSettingsMultipleDevices


class Receive(EventWithDevices):
    receiver = {
        'type': ['objectid', 'dict', 'string'],  # We should not add string but it does not work otherwise...
        'data_relation': {
            'resource': 'accounts',
            'field': '_id',
            'embeddable': True,
        },
        'schema': unregistered_user,
        'get_from_data_relation_or_create': 'email',
        'required': True,
        'sink': 2
    }
    acceptedConditions = {
        'type': 'boolean',
        'required': True,
        'allowed': {True}
    }
    type = {
        'type': 'string',
        'required': True,
        'allowed': {'FinalUser', 'CollectionPoint', 'RecyclingPoint'}
    }
    automaticallyAllocate = {
        'type': 'boolean',
        'default': False,
        'description': 'Allocates to the user'
    }
    receiverOrganization = {  # Materialization of the organization that, by the time of the receive, the user worked in
        'type': 'string',
        'readonly': True
    }
    place = place
    components = copy.deepcopy(components)


Receive.components['readonly'] = True


# receive['@type'][OR] = 'employee', ('receiver', 'unregisteredReceiver.name')
# Receiver OR ReceiverEmail. We need to hook this in a required field so it is always executed
# And @type is an always required field so we can happily hook on it


class ReceiveSettings(EventSubSettingsMultipleDevices):
    _schema = Receive
    fa = 'fa-cart-arrow-down'
    sink = -7
    short_description = 'Someone receives the devices: you, a transporter, the client...'



from bson.objectid import ObjectId
from eve.io.mongo import Mongo
from flask import current_app


class DataLayer(Mongo):
    def current_mongo_prefix(self, resource=None):
        """
        Overrides the default Eve's database selection process, by forcing the resources that use the default database
        settings, to effectively use the default database, regardless of any other setting

        By default, eve sets maximum priority to the database manually set by set_mongo_prefix when we get it from the
        URL.
        :param resource:
        :return:
        """
        if resource is None or not current_app.config['DOMAIN'][resource]['use_default_database']:
            return super(DataLayer, self).current_mongo_prefix(resource)
        else:
            return 'MONGO'

    def aggregate(self, resource, pipeline):
        datasource, *_ = self.datasource(resource)
        return list(self.pymongo(resource).db[datasource].aggregate(pipeline))

    def find_raw(self, resource, query_filter):
        datasource, *_ = self.datasource(resource)
        return list(self.pymongo(resource).db[datasource].find(query_filter))

    def find_one_raw(self, resource, id_or_query: ObjectId or dict or str):
        if type(id_or_query) is dict:
            datasource, *_ = self.datasource(resource)
            return self.pymongo(resource).db[datasource].find_one(id_or_query)
        else:
            return super(DataLayer, self).find_one_raw(resource, id_or_query)

    def drop_dtabases(self):
        self.pymongo().db.dropDatabases()



import calendar
import datetime

from flask import current_app
from pymongo import DESCENDING
from werkzeug.datastructures import ImmutableList

from ereuse_devicehub.exceptions import StandardError
from ereuse_devicehub.utils import cache


class Aggregation:
    """
    The methods of the Aggregation class return aggregated data ready to be presented with Chart.js and similar.
    """

    # For how much time is the data cached?
    CACHE_TIMEOUT = 1 * 60  # 1 minute

    def __init__(self, resouce_name):
        self.resource_name = resouce_name

    def devices_per_event_subject_month(self, options):
        """
        Returns the number of devices per event, subject and month.

        Subject is the series of the graph, and
        :param options:
        :return:
        """
        pipeline = [
            {
                '$match': {
                    '@type': options['event'],
                    '_created': {'$gte': datetime.datetime(datetime.date.today().year, 1, 1)}
                }
            },
            {
                '$unwind': '$devices'
            },
            {
                '$group': {
                    '_id': {
                        'month': {'$month': '$_created'},
                        'subject': '$' + options['subject']
                    },
                    'arrayOfDevices': {'$push': '$devices'},
                }
            },
            {
                '$project': {
                    '_id': False,
                    'month': '$_id.month',
                    'subject': '$_id.subject',
                    'countPerSubjectAndMonth': {'$size': '$arrayOfDevices'}
                }
            },
            {
                '$sort': {
                    'subject': DESCENDING,
                    'month': DESCENDING
                }
            },
            {
                '$group': {
                    '_id': {
                        'subject': '$subject'
                    },
                    'devices': {'$push': '$countPerSubjectAndMonth'},
                    'months': {'$push': '$month'}
                }
            },
            {
                '$project': {
                    '_id': False,
                    'subject': '$_id.subject',
                    'devices': True,
                    'months': True
                }
            }
        ]
        if 'receiverType' in options and options['event'] == 'Receive':
            pipeline[0]['$match']['type'] = options['receiverType']
        res = {
            'labels': list(calendar.month_name)[1:],
            'series': [],
            'data': []
        }
        a = self._aggregate(ImmutableList(pipeline))
        for org in a:
            res['series'].append('Others' if org['subject'] is None else org['subject'])
            res['data'].append([0] * len(res['labels']))
            i = 0
            for pos in org['months']:
                res['data'][-1][pos - 1] = org['devices'][i]
                i += 1
        return res

    @cache.memoize(timeout=CACHE_TIMEOUT)
    def _aggregate(self, pipeline):
        return current_app.data.aggregate(self.resource_name, pipeline)


class AggregationError(StandardError):
    status_code = 400



"""
    Execute this script in the folder of some files, and modify the list in the main method with the registered labels
    in the system.

    This script copies in the 'found' folder the snapshots which labels where found in the giving list, and in the
    'not found' folder all thte other snapshots.

    This works with the original snapshot json file.
"""
import json
import os
import shutil
from pprint import pprint

this_directory = os.path.dirname(os.path.realpath(__file__))


def get_json_from_file(filename: str) -> dict:
    with open(os.path.abspath(os.path.join(this_directory, filename))) as data_file:
        return json.load(data_file)


def main(labels):
    not_found_files = []
    found_files = []
    nf_dest_dict = os.path.join(this_directory, 'not found')
    os.makedirs(nf_dest_dict, exist_ok=True)
    f_dest_dict = os.path.join(this_directory, 'found')
    os.makedirs(f_dest_dict, exist_ok=True)
    for filename in os.listdir('.'):
        if 'json' in filename:
            event = get_json_from_file(filename)
            label = event['label']
            if int(label) not in labels:
                not_found_files.append(filename)
                shutil.copyfile(filename, os.path.join(nf_dest_dict, filename))
            else:
                found_files.append(filename)
                shutil.copyfile(filename, os.path.join(f_dest_dict, filename))
    not_found_files.sort()
    found_files.sort()
    return not_found_files, found_files


not_found, found = main(
    (1, 12, 10, 13, 14, 15, 16, 17, 18, 19, 2, 20, 21, 26, 27, 28, 3, 30, 34, 4, 42, 46, 5, 6, 7, 8, 9))
pprint("found")
pprint(found)
pprint("not found")
pprint(not_found)
pprint("Not found files copied")



import requests
from flask import json
from requests import HTTPError

from ereuse_devicehub.resources.event.device import DeviceEventDomain
from ereuse_devicehub.resources.submitter.translator import Translator
from ereuse_devicehub.rest import execute_get
from ereuse_devicehub.security.request_auth import Auth
from ereuse_devicehub.utils import Naming


class Submitter:
    """
        Submits resources to other agents.

        Submitter is thought to be working outside of Flask's application context, in another thread.
    """
    def __init__(self, token: str, app: 'DeviceHub', domain: str, translator: Translator, auth: Auth, debug=False):
        """
        :param token: Token of the Submitter user in DeviceHub.
        :param domain: The destination domain or IP.
        :param translator: A translator instance.
        :param auth: An Auth instance.
        :param debug: If true, data is not actually submitted but locally logged.
        """
        self.config = app.config
        self.domain = domain
        self.translator = translator
        self.debug = debug
        self.auth = auth
        self.token = token
        self.app = app
        self.logger = app.logger
        self.embedded = {'device': 1, 'devices': 1, 'components': 1}

    def submit(self, resource_id: str, database: str, resource_name: str):
        """
        Submits the resource to the configured agent.
        :param resource_id: The identifier (_id) in DeviceHub of the resource.
        :param database: The database or inventory (db1...) to get the resource from.
        :param resource_name: The name of the resource.
        """
        #path = self.config['DOMAIN'][resource_name]['url']
        url = '{}/{}/{}{}'.format(database, 'events', resource_id, '?embedded={}'.format(json.dumps(self.embedded)))
        with self.app.app_context():
            event = execute_get(url, self.token)
        for translated_resource, original_resource in self.translator.translate(database, event):
            submission_url = self.generate_url(original_resource, translated_resource)
            self._post(translated_resource, submission_url)

    def generate_url(self, original_resource, translated_resource) -> str:
        """Generates the url to submit the resource to, in the external agent."""
        raise NotImplementedError()

    def _post(self, resource: dict, url: str, **kwargs):
        """Sends the resource to the agent. Kwargs are sent to Request's Post"""
        if self.debug:
            self.logger.info('GRDLogger: succeed FAKE post event \n{}\n to url {}'.format(json.dumps(resource), url))
        else:
            r = requests.post(url, json=resource, auth=self.auth, **kwargs)
            try:
                r.raise_for_status()
            except HTTPError or ConnectionError:
                text = str(r.json()) if 200 <= r.status_code < 300 else ''
                error = 'Error: event \n{}\n: {} from url {} \n {}'.format(json.dumps(resource), r.status_code, url, text)
                self.logger.error(error)
            else:
                self.logger.info("GRDLogger: Succeed POST event \n{}\n from {}".format(json.dumps(resource), url))



import json

from ereuse_devicehub.exceptions import InnerRequestError
from eve.methods.delete import deleteitem_internal
from eve.methods.patch import patch_internal
from eve.methods.post import post_internal
from flask import request, current_app


def execute_post(resource: str, payload: dict, skip_validation=False):
    response = post_internal(resource, payload, skip_validation)
    if response[3] != 201:  # statusCode
        raise InnerRequestError(response[3], response[0])
    return response[0]  # Actual data


def execute_get(url: str, token: str = None):
    http_authorization = request.headers.environ['HTTP_AUTHORIZATION'] if token is None else 'Basic ' + token
    response = current_app.test_client().get(url, environ_base={'HTTP_AUTHORIZATION': http_authorization})
    data = json.loads(response.data.decode())  # It is useless to use json_util
    if response._status_code != 200:
        data['url'] = url
        raise InnerRequestError(response._status_code, data)
    else:
        return data


def execute_patch(resource: str, payload: dict, identifier):
    payload['_id'] = str(identifier)
    response = patch_internal(resource, payload, False, False, **{'_id': str(identifier)})
    if response[3] != 200:  # statusCode
        raise InnerRequestError(response[3], response[0])
    return response[0]


def execute_delete(resource: str, identifier):
    _, _, _, status = deleteitem_internal(resource, **{'_id': str(identifier)})
    if status != 204:
        raise InnerRequestError(status, {})



class BasicError(Exception):
    status_code = None

    def __init__(self, body: dict or str, status_code: int = 500):
        self.body = body
        if self.status_code is None:
            self.status_code = status_code

    def to_dict(self):
        return self.body


class StandardError(BasicError):
    message = None

    def __init__(self, message=""):
        if self.message is None:
            self.message = message

    def to_dict(self):
        return {
            '_error': {
                'message': self.message,
                'code': self.status_code,
                '@type': type(self).__name__
            },
            '_issues': {
                type(self).__name__: self.message
            },
            '_status': 'ERR'
        }


class SchemaError(StandardError):
    status_code = 422

    def __init__(self, field=None, message=None):
        if field:
            self.field = field
        if message:
            self.message = message

    def to_dict(self):
        return {
            '_issues': {
                self.field: self.message
            }
        }


class UnauthorizedToUseDatabase(StandardError):
    status_code = 401
    message = 'User has no access to this database.'


class InnerRequestError(BasicError):
    """
    Encapsulates errors produced by internal requests (GET, POST...).
    """

    def __init__(self, status_code, info: dict = None):
        self.info = info
        super().__init__(info, status_code)


class WrongCredentials(StandardError):
    """
    Error when login and user/pass do not match.
    """
    status_code = 401
    message = 'There is not an user with the matching username/password'


class RedirectToClient(Exception):
    """
    An exception that forces a redirection to the client.
    """
    pass


class RequestAnother(BasicError):
    """Redirects to DELETE Snapshot"""





from flask import request, jsonify
from passlib.handlers.sha2_crypt import sha256_crypt

from ereuse_devicehub.exceptions import WrongCredentials
from ereuse_devicehub.flask_decorators import crossdomain
from ereuse_devicehub.resources.account.domain import AccountDomain, UserNotFound


@crossdomain(origin='*', headers=['Content-Type', 'Authorization'])
def login():
    """
    Performs a login. We make this out of eve, being totally open.
    :return:
    """
    try:
        account = AccountDomain.get_one({'email': request.json['email']})
        if not sha256_crypt.verify(request.json['password'], account['password']):
            raise WrongCredentials()
        account['token'] = AccountDomain.hash_token(account['token'])
        account['_id'] = str(account['_id'])
        return jsonify(account)
    except (KeyError, TypeError, UserNotFound):
        raise WrongCredentials()



import json
import sys
from pprint import pprint

from pymongo import MongoClient

from ereuse_devicehub.exceptions import StandardError
from ereuse_devicehub.resources.account.hooks import hash_password, generate_token
from ereuse_devicehub.resources.account.domain import AccountDomain


def create_account(email: str, password: str, databases: list,
                   role: str = None, name: str = None, organization: str = None, blocked: bool = False,
                   default_database: str = None, mongo_host: str = None, mongo_port: int = None,
                   db_name: str = 'dh__accounts'):
    """
    Creates an account. You can call the file directly::

        create_account.py b@b.b 123 [\'db1\'] superuser

    :param email:
    :param password:
    :param databases: The databases the user has access to
    :param mongo_host: Leave empty to use default.
    :param mongo_port: Leave empty to use default.
    :param db_name: The name of the database to create the account to. Leave empty to use default.
    :param default_database: if none we use the first database of *databases*.
    :param role: the role of the user, if none, we use the default set in Account type.
    :param name:
    :param organization:
    :param blocked: False if it is not set
    :throws UserAlreadyExists:
    :return: The account with the *base256* token and *hashed* password.
    """
    connection = MongoClient(mongo_host, mongo_port)
    db = connection[db_name]
    if db.accounts.find_one({'email': email}):
        raise UserAlreadyExists()
    databases = eval(databases)
    if type(databases) is not list:
        raise TypeError('databases has to be a list')
    account = {
        'email': email,
        'password': password,
        'databases': databases,
        'defaultDatabase': default_database or databases[0],
        '@type': 'Account',
        'blocked': eval(blocked) if type(blocked) is str else blocked,
        'active': True
    }
    if role is not None:
        account['role'] = role
    if name is not None:
        account['name'] = name
    if organization is not None:
        account['organization'] = organization
    token = generate_token()
    while db.accounts.find_one({'token': token}) is not None:
        token = generate_token()
    account["token"] = token
    hash_password([account])
    db.accounts.insert(account)
    returned_account = db.accounts.find_one({'email': email})
    return returned_account, AccountDomain.hash_token(returned_account['token'])


class UserAlreadyExists(StandardError):
    message = 'User already exists'
    code = 309

account, hashed_token = create_account(*sys.argv[1:])
account['_id'] = str(account['_id'])
print('Account:')
print(json.dumps(account, indent=4))
print('Hashed token for REST:')
print(hashed_token)



from ereuse_devicehub.resources.device.component.settings import Component, ComponentSubSettings
from ereuse_devicehub.resources.schema import UnitCodes


class GraphicCard(Component):
    memory = {
        'type': 'float',
        'unitCode': UnitCodes.mbyte,  # MB
        'min': 1,
        'sink': 3
    }


class GraphicCardSettings(ComponentSubSettings):
    _schema = GraphicCard



from ereuse_devicehub.resources.event.settings import Event, EventSettings

prefix = {'prefix': 'devices'}


class DeviceEvent(Event):
    geo = {
        'type': 'point',
        'sink': -5,
        'description': 'Where did it happened'
        # 'anyof': [{'required': True}, {'dependencies': ['place']}]  # me OR places
    }

settings = DeviceEvent._settings.copy()
settings.update({'url': 'devices'})
settings.update(prefix)
DeviceEvent._settings = settings  # todo make this nice


class EventWithOneDevice(DeviceEvent):
    device = {
        'type': 'string',
        'data_relation': {
            'resource': 'devices',
            'field': '_id',
            'embeddable': True
        },
        'required': True
    }
EventWithOneDevice._settings = dict(Event._settings, **prefix)


class EventWithDevices(DeviceEvent):
    devices = {
        'type': 'list',
        'schema': {
            'type': 'string',
            'data_relation': {
                'resource': 'devices',
                'field': '_id',
                'embeddable': True
            }
        },
        'required': True
    }
EventWithDevices._settings = EventWithOneDevice._settings

place = {
    'type': 'objectid',
    'data_relation': {
        'resource': 'places',
        'field': '_id',
        'embeddable': True
    },
    'sink': 0,
    'description': 'Where did it happened'
}

# Materialized fields

"""
For add/remove, the user supplies this info. For allocate/locate/receive, this info is
a readonly materialization of the the components of all the computers affected, to get an easier relationship
between component - event when the event is performed to a parent
"""
components = {
    'type': 'list',
    'schema': {
        'type': 'string',
        'data_relation': {
            'resource': 'devices',
            'field': '_id',
            'embeddable': True
        }
    },
    'description': 'Components affected by the event.',
    'teaser': False
}
parent = {
    'type': 'string',
    'data_relation': {
        'resource': 'devices',
        'field': '_id',
        'embeddable': True
    },
    'description': 'The event triggered in this computer.'
}


class DeviceEventSettings(EventSettings):
    _schema = DeviceEvent


class EventSubSettings(DeviceEventSettings):
    _schema = False
    resource_methods = ['POST']
    item_methods = ['GET', 'DELETE']


class EventSubSettingsOneDevice(EventSubSettings):
    _schema = False


class EventSubSettingsMultipleDevices(EventSubSettings):
    _schema = False





import requests
from requests.auth import AuthBase


class Auth(AuthBase):
    """
    Handles the authorization method.

    If there is no available token for us, it logs-in and stores the token. Appends the token to the header accordingly.
    """

    def __init__(self, domain: str, email: str, password: str, login_path: str = 'login', auth_header_title='Basic'):
        self.domain = domain
        self.email = email
        self.password = password
        self.token = None
        self.login_path = login_path
        self.auth_header_title = auth_header_title

    def __call__(self, r):
        if self.token is None:
            self.token = self.login()
        r.headers['Authorization'] = '{} {}'.format(self.auth_header_title, self.token)
        return r

    def login(self):
        account = {
            'email': self.email,
            'password': self.password
        }
        r = requests.post('{}/{}'.format(self.domain, self.login_path), json=account)
        data = r.json()
        return data['token']



import json
from datetime import datetime

from flask import current_app
from requests import Response
from werkzeug.local import LocalProxy

from ereuse_devicehub.exceptions import RedirectToClient
from ereuse_devicehub.utils import get_header_link

def redirect_on_browser(resource, request, lookup):
    """
    Redirects the browsers to the client webApp.
    :param resource:
    :param request:
    :param lookup:
    :return:
    """
    if request.accept_mimetypes.accept_html:
        raise RedirectToClient()


def set_response_headers_and_cache(resource: str, request: LocalProxy, payload: Response):
    """
    Sets JSON Header link referring to @type
    """
    if (payload._status_code == 200 or payload._status_code == 304) and resource is not None:
        data = json.loads(payload.data.decode(payload.charset))
        resource_type = resource
        try:
            resource_type = data['@type']
        except KeyError:
            if payload._status_code == 304:
                payload.cache_control.max_age = current_app.config['ITEM_CACHE']
        else:
            # If we are here it means it is an item endpoint, not a list (resource) endpoint
            payload.cache_control.max_age = current_app.config['ITEM_CACHE']
        payload.headers._list.append(get_header_link(resource_type))


def set_date(_, resources: dict):
    """Eve's date is not precise enough (up to seconds) for massive insertion of resources."""
    for resource in resources:
        resource['_created'] = resource['_updated'] = datetime.utcnow()



from flask import current_app

from ereuse_devicehub.exceptions import StandardError
from ereuse_devicehub.resources.device.component.domain import ComponentDomain
from ereuse_devicehub.resources.device.domain import DeviceDomain
from ereuse_devicehub.resources.domain import Domain
from ereuse_devicehub.resources.place.settings import PlaceSettings


class PlaceDomain(Domain):
    resource_settings = PlaceSettings

    @staticmethod
    def get_with_coordinates(coordinates: list) -> dict:
        """
        Gets a place that intersects the given Point coordinates.
        :param coordinates: GEOJSON coordinates that represent a Point
        :return: place
        """
        place = current_app.data.find_one_raw('places', {
            'geo': {
                '$geoIntersects': {
                    '$geometry': {
                        'type': 'Point',
                        'coordinates': coordinates
                    }
                }
            }
        })
        if not place:
            raise NoPlaceForGivenCoordinates()
        return place

    @staticmethod
    def update_devices(original: set, updated: set, replaced_place_id: str or None):
        """
        Given the original set of devices of a place, and the updated version, it updates the database to achieve
        the updated set of devices.

        :param original:
        :param updated:
        :param replaced_place_id: The identifier of the replaced place. If updated is empty, this value is not used.
        :return:
        """
        devices_to_remove_id = original - updated
        children_to_remove_id = ComponentDomain.get_components_in_set(devices_to_remove_id)
        devices_to_remove_id |= children_to_remove_id

        devices_to_add_id = updated - original
        children_to_add_id = ComponentDomain.get_components_in_set(devices_to_add_id)
        devices_to_add_id |= children_to_add_id

        for device_id in devices_to_remove_id:
            PlaceDomain.device_unset_place(device_id)
        for device_id in devices_to_add_id:
            PlaceDomain.device_set_place(device_id, replaced_place_id)

    @staticmethod
    def device_set_place(device_id: str, place_id: str):
        device = DeviceDomain.get_one(device_id)
        if 'place' in device:
            current_app.data.driver.db['places'].update_one({'_id': device['place']}, {'$pull': {'devices': device_id}})
        current_app.data.driver.db['devices'].update_one({'_id': device_id}, {'$set': {'place': place_id}})

    @staticmethod
    def device_unset_place(device_id: str):
        current_app.data.driver.db['devices'].update_one({'_id': device_id}, {'$unset': {'place': ''}})


class CannotDeleteIfHasEvent(StandardError):
    message = "You cannot delete a place where you performed an event."


class NoPlaceForGivenCoordinates(StandardError):
    """
    We throw this error if given coordinates do not match any existing place.
    We just throw it in particular cases. Example: Receive and Location.
    """
    status_code = 400
    message = 'There is no place in such coordinates'


class CoordinatesAndPlaceDoNotMatch(StandardError):
    """
    Similar as NoPlaceForGivenCoordinates, this error is thrown when user supplies coordinates
    and a place, and they differ.
    """
    status_code = 400
    message = 'Place and coordinates do not match'



from ereuse_devicehub.resources.event.device import DeviceEventDomain
from ereuse_devicehub.rest import execute_post
from ereuse_devicehub.utils import Naming


class EventProcessor:
    def __init__(self):
        self.events = {}
        self.references = {}  # We cannot use device-dict as references and we cannot rely on _id or hid,
        # So we need another reference: the python's id(). One we insert the device we update the dict with _id

    def append_remove(self, component, old_parent):
        self._append(DeviceEventDomain.new_type('Remove'), old_parent, component)

    def append_add(self, component, new_parent):
        self._append(DeviceEventDomain.new_type('Add'), new_parent, component)

    def _append(self, event, common, unique):
        """
        Stores an event so it is executed later.

        :param event:
        :param common: Common property of the event (usually 'device' property).
        :param unique: Unique property of the event (usually 'component' one).
        :return:
        """
        reference = id(common)
        self.references[reference] = common
        self.events.setdefault(event, {}).setdefault(reference, []).append(unique)

    def process(self) -> list:
        """
        Executes all events stored.

        First execute the inserts so the stored devices can get the _id and then executes the rest of events.
        :return: A list of the executed events
        """
        log = []  # log done events
        for event_name, common_reference_dict in self.events.items():
            for reference, unique in common_reference_dict.items():
                device = self.references[reference]
                log.append(execute_post(Naming.resource(event_name), {
                    '@type': event_name,
                    'device': device['_id'],
                    'components': [str(x['_id']) for x in unique]
                }))
        return log



from flask import current_app as app

from ereuse_devicehub.resources.account.domain import AccountDomain

GET_PROJECTION_BLACKLIST = 'get_projection_blacklist'
GET_PROJECTION_WHITELIST = 'get_projection_whitelist'


def project(resource: str, item: dict):
    """
    Projects just the allowed fields for the actual user.

    Use this in resource settings to project just some fields accordingly to some rules and the
    role of the user.

    Both whitelist and blacklist expect a dictionary of tuples. Whitelist overrides the blacklisted
    fields, and the fields that are not shown in blacklist or whitelist are sent to user regularly.
    Every key of the dictionary is a rol or a reserved key, and the tuple is a list of affected fields.

    Example:
    In GET_PROJECTION_BLACKLIST:
    {'role1': ('field1','field2')}
    This means that role1 won't be able to get field1 or field2.

    Then, in GET_PROJECTION_WHITELIST:
    {'author': ('field1',)}
    This means that the author, regardless of the role it has (even if it has role1) will retreive
    field1.

    Reserved keys for blacklist:
    - *: Wildcard. It means all roles.

    Reserved keys for whitelist:
    - author: When scenarios where just the author can get a field. For example, it's password.
            Author is computed using the _id when resource is account or byUser otherwise.
    """

    delete_fields = []
    if 'get_projection_blacklist' in app.config['DOMAIN'][resource]:  # We take note of the blacklisted fields
        blacklist = app.config['DOMAIN'][resource]['get_projection_blacklist']
        for role, fields in blacklist.items():
            if '*' == role or AccountDomain.actual['role'] <= role:  # Comparing a role means comparing its permission grade
                delete_fields += fields
    if 'get_projection_whitelist' in app.config['DOMAIN'][resource]:  # We remove from the blacklist the white ones
        for role, fields in app.config['DOMAIN'][resource]['get_projection_whitelist'].items():
            if role == 'author':
                if (resource == 'account' and AccountDomain.actual['_id'] == item['_id']) or ('byUser' in item):
                    delete_fields -= fields
            elif AccountDomain.actual.role >= role:
                delete_fields -= fields
    for field in delete_fields:  # We delete the blacklisted fields
        if field in item:
            del item[field]



from ereuse_devicehub.resources.device.schema import Device
from ereuse_devicehub.resources.device.settings import DeviceSubSettings


class Computer(Device):
    type = {
        'type': 'string',
        'allowed': {'Desktop', 'Laptop', 'Netbook', 'Server', 'Microtower'}
    }
    forceCreation = {
        'type': 'boolean',
        'default': False
    }


class ComputerSettings(DeviceSubSettings):
    _schema = Computer
    etag_ignore_fields = DeviceSubSettings.etag_ignore_fields + ['forceCreation']



from assertpy import assert_that
from ereuse_devicehub.tests.test_resources.test_events.test_device_event import TestDeviceEvent


class TestLocate(TestDeviceEvent):
    LOCATE = 'locate'
    POST_LOCATE = TestDeviceEvent.DEVICE_EVENT + '/locate'

    def test_create_locate_with_place(self):
        locate = self.get_fixture(self.LOCATE, 'locate')
        locate['place'] = self.place['_id']
        locate['devices'] = self.devices_id
        locate = self.post_and_check(self.POST_LOCATE, locate)
        # Let's check the place has been correctly materialized in the devices
        for device_id in self.devices_id:
            self.device_and_place_contain_each_other(device_id, self.place['_id'])
        return locate

    def test_delete(self):
        locate = self.test_create_locate_with_place()
        self.delete_and_check(self.POST_LOCATE + '/' + locate['_id'])
        _, status = self.get(self.LOCATE, '', locate['_id'])
        self.assert404(status)
        # let's check the place has been de-materialized in the devices
        for device_id in self.devices_id:
            self.devices_do_not_contain_places(device_id)
        # Let's check locate has been de-materialized in the devices
        for device_id in self.devices_id:
            device, _ = self.get(self.DEVICES, '', device_id)
            assert_that(device['events']).extracting('_id').does_not_contain(locate['_id'])

    def test_create_locate_with_coordinates(self):
        # Let's create a location with coordinates without a place that contains them. It will give us error.
        locate = self.get_fixture(self.LOCATE, 'locate_with_coordinates')
        locate['devices'] = self.devices_id
        try:
            self.post_and_check(self.POST_LOCATE, locate)
        except AssertionError as e:
            if e.args[0] == '400 != 201':
                # Now we create the place
                place = self.post_fixture(self.PLACES, self.PLACES, 'place_with_coordinates')
                # We repeat the location. As the coordinates are in the area of the place, we can create it
                locate = self.post_and_check(self.POST_LOCATE, locate)
                # Let's check if locate has been assigned to the place
                locate, _ = self.get(self.DEVICE_EVENT, '', locate['_id'])
                self.assertIn('place', locate)
                self.assertEqual(locate['place'], place['_id'])
                # Let's assure that the materializations of the devices are correct
                for device_id in self.devices_id:
                    device, _ = self.get(self.DEVICES, '', device_id)
                    self.assertEqual(place['_id'], device['place'])
                    assert_that(device['events'][0]).is_subset_of(locate)
            else:
                self.assertTrue(False)
        else:
            self.assertTrue(False)

    def test_monitor(self):
        """
        The same as test_create_locate_with_coordiantes but with a monitor
        """
        locate = self.get_fixture(self.LOCATE, 'locate_with_coordinates')
        snapshot = self.post_fixture(self.SNAPSHOT, '{}/{}'.format(self.DEVICE_EVENT, self.SNAPSHOT), 'monitor')
        snapshot, _ = self.get('devices_snapshot', '', snapshot['_id'])
        locate['devices'] = [self.get(self.DEVICES, '', snapshot['device'])[0]['_id']]
        # Let's directly create a place
        place = self.post_fixture(self.PLACES, self.PLACES, 'place_with_coordinates')
        # We repeat the location. As the coordinates are in the area of the place, we can create it
        locate = self.post_and_check(self.POST_LOCATE, locate)
        # Let's check if locate has been assigned to the place
        locate, _ = self.get(self.DEVICE_EVENT, '', locate['_id'])
        self.assertIn('place', locate)
        self.assertEqual(locate['place'], place['_id'])
        # Let's assure that the materializations of computerMonitor are correct
        computerMonitor, _ = self.get(self.DEVICES, '', snapshot['device'])
        self.assertEqual(place['_id'], computerMonitor['place'])
        assert_that(computerMonitor['events'][0]).is_subset_of(locate)



import pymongo
from ereuse_devicehub.resources.account.domain import AccountDomain, UserNotFound
from ereuse_devicehub.resources.device.domain import DeviceDomain
from ereuse_devicehub.resources.event.device.deallocate.deallocate import AlreadyAllocated


def materialize_actual_owners_add(allocates: list):
    for allocate in allocates:
        properties = {'$addToSet': {'owners': allocate['to']}}
        DeviceDomain.update_raw(allocate['devices'], properties)
        DeviceDomain.update_raw(allocate.get('components', []), properties)


def avoid_repeating_allocations(allocates: list):
    """
    Checks that we are not allocating to an account that is already an owner of the device

    This method must execute after
    :param allocates:
    :return:
    """
    for allocate in allocates:
        devices_with_repeating_owners = DeviceDomain.get({
            '$or': [
                {'_id': {'$in': [allocate['devices']]}},
                {'owners': {'$in': [allocate['to']]}}
            ]
        })
        ids = [device['_id'] for device in devices_with_repeating_owners]
        allocate['devices'] = list(set(allocate['devices']) - set(ids))
        if len(allocate['devices']) == 0:
            raise AlreadyAllocated()


def set_organization(allocates: list):
    for allocate in allocates:
        try:
            allocate['toOrganization'] = AccountDomain.get_one(allocate['to'])['organization']
        except (UserNotFound, KeyError):
            pass


def re_materialize_owners(_, event: dict):
    """Deleting an Allocate / deallocate requires """
    if event.get('@type') in ('devices:Allocate', 'devices:Deallocate'):
        from ereuse_devicehub.resources.event.device.allocate.allocate import materialize_owners
        materialize_owners(event['devices'])



from flask import Request, g
from flask import json, current_app
from flask.wrappers import _missing, _get_data

from ereuse_devicehub.exceptions import StandardError

PGP_MESSAGE = '-----BEGIN PGP SIGNED MESSAGE-----'


# noinspection PyAttributeOutsideInit
class RequestSignedJson(Request):
    """
        Adds the ability to parse signed JSON.

        It sets g.trusted_json to true if the json was correctly signed (verified), false otherwise.
    """

    def get_json(self, force=False, silent=False, cache=True):
        rv = getattr(self, '_cached_json', _missing)
        if rv is not _missing:
            return rv
        json_data = self.get_signed_json(cache)
        if json_data is None:
            json_data = super().get_json(force, silent, cache)
            g.trusted_json = False
        return json_data

    def get_signed_json(self, cache=True):
        """
        Verifies and returns a signed JSON. This code is strongly inspired by `flask.Request.get_json`

        :param cache:
        :return:
        """
        request_charset = self.mimetype_params.get('charset')
        encrypted_data = _get_data(self, cache)
        if not encrypted_data.startswith(PGP_MESSAGE.encode()):
            return None

        current_app.gpg.encoding = request_charset if request_charset is not None else 'utf-8'
        if not current_app.gpg.verify(encrypted_data):
            raise SignatureCannotBeVerified()
        decrypted_data = str(current_app.gpg.decrypt(encrypted_data))

        try:
            if request_charset is not None:
                rv = json.loads(decrypted_data, encoding=request_charset)
            else:
                rv = json.loads(decrypted_data)
        except ValueError as e:
            rv = self.on_json_loading_failed(e)
        if cache:
            self._cached_json = rv
        g.trusted_json = True
        return rv


class SignatureCannotBeVerified(StandardError):
    status_code = 400
    message = "The signature of the file cannot be verified. Have you uploaded the Public key?"



from ereuse_devicehub.resources.device.benchmark_settings import BenchmarkHardDrive
from ereuse_devicehub.resources.device.component.settings import Component, ComponentSubSettings
from ereuse_devicehub.resources.event.device.erase_basic.settings import EraseBasic
from ereuse_devicehub.resources.event.device.test_hard_drive.settings import TestHardDrive
from ereuse_devicehub.resources.schema import UnitCodes


class HardDrive(Component):
    interface = {
        'type': 'string',
        'sink': -1,
        'teaser': False
    }
    size = {
        'type': 'float',
        'unitCode': UnitCodes.mbyte,
        'sink': 1
    }
    erasure = {
        'type': 'dict',
        'schema': EraseBasic,
        'writeonly': True
    }
    erasures = {
        'type': 'list',
        'schema': {
            'type': 'objectid',
            'data_relation': {
                'resource': 'events',
                'field': '_id',
                'embeddable': True
            }
        },
        'readonly': True
    }
    firmwareRevision = {
        'type': 'string',
        'teaser': False,
        'sink': -1
    }
    blockSize = {
        'type': 'integer',
        'sink': -1,
        'teaser': False
    }
    sectors = {
        'type': 'integer',
        'sink': -1,
        'teaser': False
    }
    test = {
        'type': 'dict',
        'schema': TestHardDrive
    }
    tests = {
        'type': 'list',
        'schema': {
            'type': 'objectid',
            'data_relation': {
                'resource': 'events',
                'field': '_id',
                'embeddable': True
            }
        },
        'readonly': True
    }
    benchmark = {
        'type': 'dict',
        'schema': BenchmarkHardDrive,
        'writeonly': True
    }
    benchmarks = {
        'type': 'list',
        'schema': {
            'type': 'dict',
            'schema': BenchmarkHardDrive
        },
        'readonly': True
    }

    @classmethod
    def actual_fields(cls):
        fields = super().actual_fields()
        if cls._import_schemas:
            # todo We need to add this because these events are taken from hardDrive in snapshot,
            # We should remove them in Device POST, so they don't pass through the API and get generated.
            # This fields are generated automatically by the API
            del fields['erasure']['schema']['incidence']['default']
            del fields['erasure']['schema']['secured']['default']
            del fields['test']['schema']['incidence']['default']
            del fields['test']['schema']['secured']['default']
            fields['test']['schema']['device']['required'] = False
            fields['erasure']['schema']['device']['required'] = False
        return fields


class HardDriveSettings(ComponentSubSettings):
    _schema = HardDrive
    etag_ignore_fields = ComponentSubSettings.etag_ignore_fields + ['tests', 'erasures', 'test', 'erasure']



class Component:
    @staticmethod
    def get_types_of_components() -> ():
        return 'GraphicCard', 'HardDrive', 'Motherboard', 'NetworkAdapter', 'Processor', 'RamModule', 'SoundCard', 'OpticalDrive'



def hooks(app):
    """
        This method "ties" all the hooks DeviceHub uses with the app.
    """
    from ereuse_devicehub.resources.hooks import set_response_headers_and_cache, redirect_on_browser
    app.on_post_GET += set_response_headers_and_cache
    app.on_pre_GET += redirect_on_browser

    from ereuse_devicehub.security.hooks import project_item, project_resource, authorize_public, deny_public
    app.on_fetched_item += authorize_public
    app.on_fetched_resource += deny_public
    app.on_fetched_item += project_item
    app.on_fetched_resource += project_resource

    from ereuse_devicehub.resources.device.hooks import generate_etag, autoincrement, post_benchmark, \
        materialize_public_in_components, materialize_public_in_components_update
    app.on_insert += generate_etag
    app.on_insert += autoincrement
    app.on_insert += post_benchmark
    app.on_inserted += materialize_public_in_components
    app.on_updated += materialize_public_in_components_update

    from ereuse_devicehub.resources.event.device.snapshot.hooks import on_insert_snapshot, save_request, \
        materialize_test_hard_drives, materialize_erase_basic, set_secured, delete_events
    app.on_insert_devices_snapshot += set_secured
    app.on_insert_devices_snapshot += on_insert_snapshot
    app.on_insert_devices_snapshot += save_request
    app.on_inserted_devices_snapshot += materialize_test_hard_drives
    app.on_inserted_devices_snapshot += materialize_erase_basic
    app.on_delete_item += delete_events

    from ereuse_devicehub.resources.event.device.hooks import get_place, materialize_components, materialize_parent, \
        set_place, unset_place, delete_events_in_device, remove_from_other_events
    app.on_insert += get_place
    app.on_inserted += set_place
    app.on_delete_item += unset_place
    app.on_insert += materialize_components
    app.on_insert += materialize_parent
    app.on_delete_item += delete_events_in_device
    app.on_delete_item += remove_from_other_events

    from ereuse_devicehub.resources.event.device.add.hooks import add_components, delete_components
    app.on_inserted_devices_add += add_components
    app.on_delete_item += delete_components

    from ereuse_devicehub.resources.event.device.register.hooks import post_devices, delete_device
    app.on_insert_devices_register += post_devices
    app.on_delete_item += delete_device

    from ereuse_devicehub.resources.event.device.remove.hooks import remove_components
    app.on_inserted_devices_remove += remove_components

    from ereuse_devicehub.resources.account.hooks import set_byUser, set_byOrganization, \
        add_or_get_inactive_account_allocate, add_or_get_inactive_account_receive, add_or_get_inactive_account_snapshot
    app.on_insert += set_byUser
    app.on_insert_devices_receive += add_or_get_inactive_account_receive
    app.on_insert_devices_allocate += add_or_get_inactive_account_allocate
    app.on_insert_devices_snapshot += add_or_get_inactive_account_snapshot
    app.on_insert += set_byOrganization

    from ereuse_devicehub.resources.event.device.allocate.hooks import avoid_repeating_allocations, \
        materialize_actual_owners_add, set_organization, re_materialize_owners
    app.on_insert_devices_allocate += avoid_repeating_allocations
    app.on_inserted_devices_allocate += materialize_actual_owners_add
    app.on_insert_devices_allocate += set_organization
    app.on_deleted_item += re_materialize_owners

    from ereuse_devicehub.resources.event.device.deallocate.hooks import materialize_actual_owners_remove, \
        set_organization
    app.on_inserted_devices_deallocate += materialize_actual_owners_remove
    app.on_insert_devices_deallocate += set_organization

    from ereuse_devicehub.resources.submitter.grd_submitter.hooks import submit_events_to_grd
    app.on_inserted += submit_events_to_grd

    from ereuse_devicehub.resources.account.hooks import add_token, hash_password, set_default_database_if_empty
    app.on_insert_accounts += add_token
    app.on_insert_accounts += hash_password
    app.on_insert_accounts += set_default_database_if_empty

    from ereuse_devicehub.resources.event.device.receive.hooks import transfer_property, set_organization
    app.on_insert_devices_receive += transfer_property
    app.on_insert_devices_receive += set_organization

    from ereuse_devicehub.resources.place.hooks import set_place_in_devices, update_place_in_devices, \
        unset_place_in_devices, update_place_in_devices_if_places, avoid_deleting_if_has_event
    app.on_inserted_places += set_place_in_devices
    app.on_updated_places += update_place_in_devices_if_places
    app.on_delete_item_places += avoid_deleting_if_has_event
    app.on_deleted_item_places += unset_place_in_devices
    app.on_replaced_places += update_place_in_devices

    # Device materializations
    from ereuse_devicehub.resources.device.hooks import MaterializeEvents, redirect_to_first_snapshot
    app.on_inserted += MaterializeEvents.materialize_events
    app.on_delete_item += MaterializeEvents.dematerialize_event
    app.on_pre_DELETE += redirect_to_first_snapshot

    from ereuse_devicehub.resources.hooks import set_date
    app.on_insert += set_date



from time import sleep

from ereuse_devicehub.tests import TestStandard


class TestEvent(TestStandard):
    def setUp(self, settings_file=None, url_converters=None):
        super(TestEvent, self).setUp(settings_file, url_converters)
        self.place = self.post_fixture(self.PLACES, self.PLACES, 'place')
        self.devices_id = self.get_fixtures_computers()

    def tearDown(self):
        sleep(2)
        super(TestEvent, self).tearDown()



from .graphic_card import settings
from .hard_drive import settings
from .motherboard import settings
from .network_adapter import settings
from .optical_drive import settings
from .processor import settings
from .ram_module import settings
from .sound_card import settings



from ereuse_devicehub.resources.account.domain import AccountDomain, UserNotFound
from ereuse_devicehub.resources.device.domain import DeviceDomain


def materialize_actual_owners_remove(events: list):
    for event in events:
        properties = {'$pull': {'owners': event['from']}}
        DeviceDomain.update_raw(event.get('components', []), properties)
        return DeviceDomain.update_raw(event['devices'], properties)


def set_organization(deallocates: list):
    for deallocate in deallocates:
        try:
            deallocate['fromOrganization'] = AccountDomain.get_one(deallocate['from'])['organization']
        except (UserNotFound, KeyError):  # todo ensure organization is not always needed
            pass



import copy

from ereuse_devicehub.tests import TestStandard


class TestPlace(TestStandard):
    def setUp(self, settings_file=None, url_converters=None):
        super(TestPlace, self).setUp(settings_file, url_converters)
        self.place = self.get_fixture(self.PLACES, 'place')

    def device_and_place_do_not_contain_each_other(self, device_id, place_id):
        """
        Exactly opposite of :func:`device_and_place_contain_each_other`
        :param device_id:
        :param place_id:
        :return:
        """
        try:
            place, _ = self.get(self.PLACES, place_id)
            try:
                self.assertNotIn(device_id, place['devices'])
            except KeyError:
                pass
        except AssertionError:
            pass  # It is fine if the place does not exist
        device, _ = self.get(self.DEVICES, device_id)
        try:
            self.assertNotIn(place_id, device['place'])
        except KeyError:
            pass
        if 'components' in device:
            for component_id in device['components']:
                component, _ = self.get(self.DEVICES, component_id)
                try:
                    self.assertNotIn(place_id, component['place'])
                except KeyError:
                    pass

    def test_create_place_with_coordinates(self):
        self.post_and_check(self.PLACES, self.place)

    def test_place_with_devices(self):
        """
        CRUD of a place adding and removing devices through all the different methods (POST / PUT / PATCH).
        :return:
        """
        # Let's create a place with one device in it
        computers_id = self.get_fixtures_computers()
        self.place['devices'] = [computers_id[0]]
        place = self.post_and_check(self.PLACES, copy.deepcopy(self.place))
        self.device_and_place_contain_each_other(computers_id[0], place['_id'])
        # Let's PATCH de place with the device 1 and 2
        patched_place = {
            '_id': place['_id'],
            '@type': 'Place',
            'devices': place['devices'] + computers_id[1:2]
        }
        self.patch_and_check('{}/{}'.format(self.PLACES, place['_id']), patched_place)
        # Now we have computers 0, 1 and 2
        for computer_id in computers_id[:2]:
            self.device_and_place_contain_each_other(computer_id, place['_id'])
        # Let's PUT adding the last device
        place = copy.deepcopy(self.place)  # We do not want any readonly values
        place['_id'] = patched_place['_id']
        place['devices'] = computers_id
        self.put('{}/{}'.format(self.PLACES, place['_id']), place)
        for computer_id in computers_id:
            self.device_and_place_contain_each_other(computer_id, place['_id'])
        # Now let's remove a device from the place
        del place['devices'][0]
        for computer_id in computers_id[1:]:
            self.device_and_place_contain_each_other(computer_id, place['_id'])
        self.device_and_place_do_not_contain_each_other(computers_id[0], place['_id'])
        # Finally, let's remove the entire place. Devices must loose their place
        self.delete('{}/{}'.format(self.PLACES, place['_id']))
        for computer_id in computers_id:
            self.device_and_place_do_not_contain_each_other(computer_id, place['_id'])



from ereuse_devicehub.resources.event.device.settings import parent, EventWithOneDevice, EventSubSettingsOneDevice


class TestHardDrive(EventWithOneDevice):
    type = {
        'type': 'string',
        #'allowed': ['Short Offline', 'Extended Offline']
    }
    status = {
        'type': 'string',
        'required': True
    }
    lifetime = {
        'type': 'integer',
    }
    firstError = {
        'type': 'integer',
        'nullable': True
    }
    snapshot = {
        'type': 'objectid',
        'data_relation': {
            'resource': 'events',
            'field': '_id',
            'embeddable': True
        }
    }
    error = {
        'type': 'boolean',
        'required': True,
        # IF_VALUE_REQUIRE: (False, ('type', 'lifetime', 'firstError')) todo this can just be done when hard-drive is not
        # nested, so we will be able to activate it when, in snapshot, we can abort TestHardDrive event
    }
    parent = parent


class TestHardDriveSettings(EventSubSettingsOneDevice):
    _schema = TestHardDrive
    fa = 'fa-flask'
    short_description = 'A test of the health of the hard drive'
    item_methods = ['GET']



import copy
from distutils import version

import validators
from bson import json_util, ObjectId
from bson.errors import InvalidId
from cerberus import errors
from ereuse_devicehub.resources.account.role import Role
from ereuse_devicehub.utils import Naming, coerce_type
from eve.io.mongo import Validator
from eve.utils import config
from flask import current_app as app

from . import errors as dh_errors

ALLOWED_WRITE_ROLES = 'dh_allowed_write_roles'
DEFAULT_AUTHOR = 'dh_default_author'
IF_VALUE_REQUIRE = 'dh_if_value_require'
COERCE_WITH_CONTEXT = 'coerce_with_context'

HID_REGEX = '[\w]+-[\w]+-[\w]+'


class DeviceHubValidator(Validator):
    SCALE_AD = ['A', 'B', 'C', 'D']
    special_rules = Validator.special_rules + ('or', COERCE_WITH_CONTEXT, 'move')

    def __init__(self, schema=None, resource=None, allow_unknown=False, transparent_schema_rules=False):
        self._validations = {}
        """Fields that have been already validated"""
        super().__init__(schema, resource, allow_unknown, transparent_schema_rules)

    def _validate(self, document, schema=None, update=False, context=None):
        self._coerce_type(document)
        super(DeviceHubValidator, self)._validate(document, schema, update, context)
        self._validate_or(self._current)
        return len(self._errors) == 0

    """
       Removes a null field (as they equal 'undefined' ones)
    """

    def _validate_definition(self, definition, field, value):
        self._validations[field] = True
        if value is None:
            # We change this from Python-eve. For us, null values equal undefined fields and they are silently removed
            del self._current[field]
            return
        if 'move' in definition:
            self._move(definition['move'], value, field, definition)
            return  # The actual field has been moved, so there is no value any more to validate
        if COERCE_WITH_CONTEXT in definition:
            value = self._validate_coerce_with_context(definition['coerce_with_context'], field, value)
            self._current[field] = value
        super()._validate_definition(definition, field, value)

    def _validate_coerce_with_context(self, coerce, field, value):
        """Like coerce, but adds parameters to know the context, like the field and the full document."""
        try:
            value = coerce(value, field, self._current, self.schema)
        except (TypeError, ValueError):
            self._error(field, errors.ERROR_COERCION_FAILED.format(field))
        return value

    def _move(self, to, value, field, definition):
        """
        Moves one field to another one.

        :param to: Destination field. It can be 'readonly', as it will be ignored.
        """
        self._current[to] = value
        if self._validations.get(field) or field not in self.document:  # If the field was already validated
            other_definition = copy.deepcopy(self.schema.get(to))  # We do not affect original definition
            other_definition.pop('readonly', None)  # Otherwise we won't be able to 'paste' it
            self._validate_definition(other_definition, to, value)  # As the value changed we need to re-do it
        del self._current[field]

    def _validate_dh_allowed_write_roles(self, roles, field, value):
        from ereuse_devicehub.resources.account.domain import AccountDomain
        if not AccountDomain.actual['role'].has_role(roles):
            self._error(field, json_util.dumps({'ForbiddenToWrite': self.document}))

    def _validate_or(self, document):
        for field_name, definition in self.schema.items():
            if 'or' in definition:
                field_names = set([field_name] + definition['or'])
                if field_names.isdisjoint(document.keys()):
                    self._error(next(iter(field_names)),
                                'You need at least one of the following: {}'.format(field_names))

    def _coerce_type(self, fields):
        """
         A coerce method masked in a validation one, as coerce has some bugs in Cerberus 0.96.

         Warning: Do not read from @type in other validations as this method changes the value.
         """
        # todo move it to a coerce method when Cerberus 0.10 is out
        if fields:
            coerce_type(fields)

    def _validate_dh_if_value_require(self, condition: tuple, field: str, value):
        desired_value, fields = condition
        if value == desired_value:
            if not all(other_field in self.document.keys() for other_field in fields):
                self._error(field, "When {} is {}, you need to send: {}".format(field, desired_value, fields))

    def _get_resource(self, unique, field, value, query):
        if unique and self.resource:
            query[field] = value

            # exclude current document
            if self._id:
                id_field = config.DOMAIN[self.resource]['id_field']
                query[id_field] = {'$ne': self._id}

            return app.data.find_one_raw(self.resource, query)
        else:
            return None

    def _is_value_unique(self, unique, field, value, query):
        """
        We override this method to give the ability to show the _id when unique constraint fails.

        We add self.resource validation, as when in used in a dict, unique doesn't work (it is under other resource,
        anyway, so we do not expect it to work)

        We move the logic of unique in the following method.
        """
        response = self._get_resource(unique, field, value, query)
        if response:
            self._error(field, json_util.dumps({'NotUnique': response}))

    def _validate_type_hid(self, field, value):
        """
        General validation for inserting devices (the name of the function is forced by Cerberus, not a good one).

        - Tries to create hid and validates it.
        - If hid cannot be created:
            - If it has a parent, ensures that the device is unique.
            - If it has not a parent, validates that the device has an user provided _id.
        """
        from ereuse_devicehub.resources.device.component.domain import ComponentDomain
        from ereuse_devicehub.resources.device.exceptions import DeviceNotFound
        try:
            self.document['hid'] = Naming.url_word(self.document['manufacturer']) + \
                                   '-' + Naming.url_word(self.document['serialNumber']) + \
                                   '-' + Naming.url_word(self.document['model'])
        except KeyError:
            del self.document['hid']
            self.document['isUidSecured'] = False
            if '_id' not in self.document:  # We do not validate here the unique constraint of _id
                if 'parent' in self.document:
                    try:
                        component = ComponentDomain.get_similar_component(self.document, self.document['parent'])
                        self._error('model', json_util.dumps({'NotUnique': component}))
                    except (KeyError, DeviceNotFound):
                        pass
                else:
                    # If device has no parent and no hid, user needs to: or provide _id or forcing creating it
                    if 'forceCreation' not in self.document or not self.document['forceCreation']:
                        self._error('_id', json_util.dumps({'NeedsId': self.document}))
                        # else: user forces us to create the device, it will be assigned an _id
                        # else: user provided _id. We accept this, however is unsecured.
        else:
            self._validate_regex(HID_REGEX, field, self.document['hid'])
            self._validate_unique(True, field, self.document['hid'])

    def _validate_type_databases(self, field, databases):
        """Databases are a unique list of values (a set without 'set' for mongo/json compatibilities). Admins can only
        create accounts within their databases."""
        self._validate_type_list(field, databases)
        if len(databases) != len(set(databases)):
            self._error(field, json_util.dumps({'DuplicatedDatabases': 'Databases are duplicated'}))
        from ereuse_devicehub.resources.account.domain import AccountDomain
        if AccountDomain.actual['role'] < Role.SUPERUSER and not set(databases).issubset(
                set(AccountDomain.actual['databases'])):
            self._error(field, json_util.dumps(dh_errors.not_enough_privilege))

    def _validate_in_database(self, do: bool, field, identifier: ObjectId):
        """
        Validates that the input account has at least one database in common with the actual account
        :param do: Boolean to execute the method
        :param identifier: The identifier of the account
        """
        if do:
            from ereuse_devicehub.resources.account.domain import AccountDomain, UserNotFound
            try:
                # We select an user with at
                dbs = AccountDomain.actual['databases']
                AccountDomain.get_one({'_id': ObjectId(identifier), 'databases': {'$in': dbs}})
            except UserNotFound:
                self._error(field, json_util.dumps(dh_errors.not_enough_privilege))

    def _validate_data_relation(self, data_relation, field, value):
        if not isinstance(value, dict) and not isinstance(value, list):  # todo more broad way?
            super(DeviceHubValidator, self)._validate_data_relation(data_relation, field, value)

    def _validate_device_id(self, validate, field, value):
        if validate and self.resource == 'computer':
            if self._get_resource(True, field, value, {}) is None:
                self._error(field, json_util.dumps({'CannotCreateId': self.document}))

    def _validate_type_natural(self, field, value):
        self._validate_type_integer(field, value)
        if value < 0:
            self._error(field, errors.ERROR_BAD_TYPE.format('natural (positive integer)'))

    def _validate_type_url(self, field, value):
        if not validators.url(value):
            self._error(field, errors.ERROR_BAD_TYPE.format('email'))

    def _validate_type_email(self, field, value):
        if not validators.email(value):
            self._error(field, errors.ERROR_BAD_TYPE.format('email'))

    def _validate_type_version(self, field, value):
        try:
            version.StrictVersion(value)
        except ValueError:
            self._error(field, '{} is not a valid python strict version.'.format(value))

    def _validate_sink(self, nothing, field, value):
        pass

    def _validate_description(self, nothing, field, value):
        pass

    def _validate_unitCode(self, nothing, field, value):
        pass

    def _validate_doc(self, nothing, field, value):
        pass

    def _validate_get_from_data_relation_or_create(self, nothing, field, value):
        """
        Python-eve is incapable of serializing to objectid when type==[objectid, dict], rejecting the entering string.
        To make it work we add [objectid, dict, string] so the entering string is accepted and then we perform
        the conversion ourselves here.
        """
        if type(value) is not dict:
            try:
                self._current[field] = ObjectId(value)
            except InvalidId as error:
                self._error(field, str(error))

    def _validate_label(self, nothing, field, value):
        pass

    @staticmethod
    def _validate_writeonly(x, y, z):
        """
        Don't expect to GET this value.
        """
        pass

    def _validate_teaser(self, x, y, z):
        pass

    def _validate_excludes(self, other_field: list, field: str, value):
        """
        "Field A and field B cannot have both values."

        A field that is using the default value is treated as it is not using value at all. This method
        presupposes that values cannot be None.
        :param other_field:
        :param field:
        :param value:
        :return:
        """
        if other_field in self.document:
            if self.document[other_field] != self.schema[other_field].get('default', None) \
                    and value != self.schema[field].get('default', None):
                self._error(field, 'Cannot be with {} field.'.format(other_field))

    def _validate_unique_values(self, boolean, field, value):
        if boolean:
            if len(value) != len(set(value)):
                self._error(field, 'There cannot be repetitions')

    def _validate_modifiable(self, boolean, field, value):
        """
        Validates that a value is not modified: once the value has ben set, it cannot be changed.
        """
        if not boolean:
            if hasattr(self, '_original_document') and self._original_document is not None \
                    and field in self._original_document and value != self._original_document[field]:
                self._error(field, 'You cannot modify this value.')

    def _validate_materialized(self, _, field, value):
        """
        Just to show which values are materialized. They behave like *readonly*.
        """
        self._validate_readonly(True, field, value)



from ereuse_devicehub.resources.event.device.settings import place, EventWithOneDevice, EventSubSettingsOneDevice


class Register(EventWithOneDevice):
    device = {
        'type': ['dict', 'string'],  # POST dict, GET str
        'data_relation': {
            'resource': 'devices',
            'field': '_id',
            'embeddable': True
        }
    }
    components = {
        'type': ['list', 'string'],  # POST dict, GET str
        'data_relation': {
            'resource': 'devices',
            'field': '_id',
            'embeddable': True
        }
    }
    force = {
        'type': ['boolean']  # Creates a device even if it does not have pid or hid, doesn't affect components
        # An automatic way of generating pid must be set (ex: PID_AS_AUTOINCREMENT)
    }
    place = place


class RegisterSettings(EventSubSettingsOneDevice):
    _schema = Register
    extra_response_fields = EventSubSettingsOneDevice.extra_response_fields + ['device', 'components']
    fa = 'fa-plus'
    short_description = 'The creation of a new device in the system.'



import pymongo

from ereuse_devicehub.resources.device.schema import Device
from ereuse_devicehub.resources.resource import ResourceSettings
from ereuse_devicehub.validation.validation import HID_REGEX


class DeviceSettings(ResourceSettings):
    resource_methods = ['GET']
    item_methods = ['GET', 'PATCH', 'DELETE']
    _schema = Device
    additional_lookup = {
        'field': 'hid',
        'url': 'regex("' + HID_REGEX + '")'
    }
    item_url = 'regex("[\w]+")'
    mongo_indexes = {
        '@type': [('@type', pymongo.DESCENDING)],
        '@type and subtype': [('@type', pymongo.DESCENDING), ('type', pymongo.DESCENDING)],
        '@type and _created': [('@type', pymongo.DESCENDING), ('_created', pymongo.DESCENDING)]
    }
    etag_ignore_fields = ['hid', '_id', 'components', 'isUidSecured', '_created', '_updated', '_etag', 'speed',
                          'busClock', 'labelId', 'owners', 'place', 'benchmark', 'benchmarks', 'public', '_links',
                          'forceCreation', 'parent', 'events']
    cache_control = 'max-age=1, must-revalidate'
    extra_response_fields = ResourceSettings.extra_response_fields + ['hid', 'pid']
    datasource = {
        'source': 'devices'
    }
    icon = 'devices/icons/'


class DeviceSubSettings(DeviceSettings):
    _schema = False
    resource_methods = ['GET']
    item_methods = ['DELETE', 'GET']



from bson.objectid import ObjectId
from ereuse_devicehub.exceptions import StandardError
from ereuse_devicehub.resources.resource import ResourceSettings
from ereuse_devicehub.utils import ClassProperty
from flask import current_app


class Domain:
    """Provides utility methods to work with resources, like easier access to data-layers. Extend it happily."""
    resource_settings = ResourceSettings
    """Link to the resource settings. Override it linking it with the appropriate subclass"""

    @classmethod
    def get_one(cls, id_or_filter: dict or ObjectId or str):
        """
        Obtains a resource.
        :param id_or_filter: An identifier or the filter in Mongo to obtain the resource.
        :throws ResourceNotFound:
        :return:
        """
        resource = current_app.data.find_one_raw(cls.source, id_or_filter)
        if resource is None:
            raise ResourceNotFound()
        else:
            return resource

    @classmethod
    def get(cls, query_filter: dict) -> list:
        """
        Obtains several resources.
        :param query_filter: A Mongo filter to obtain the resources.
        """
        return list(current_app.data.find_raw(cls.source, query_filter))

    @classmethod
    def update_raw(cls, ids: str or ObjectId or list, operation: dict):
        """
        Sets the properties of a resource using directly the database layer.
        :param ids:
        :param operation: MongoDB update query.
        :return The number of files edited in total
        """
        resources_id = [ids] if type(ids) is str or type(ids) is ObjectId else ids
        count = 0
        for resource_id in resources_id:
            count += current_app.data.driver.db[cls.source].update_one({'_id': resource_id}, operation).modified_count
        return count

    @classmethod
    def update_many_raw(cls, filter, operation):
        return current_app.data.driver.db[cls.source].update_many(filter, operation)

    @classmethod
    def delete(cls, query):
        return current_app.data.driver.db[cls.source].delete_one(query)

    # noinspection PyNestedDecorators
    @ClassProperty
    @classmethod
    def source(cls):
        try:
            return cls.resource_settings.datasource['source']
        except AttributeError:
            raise AttributeError('Make sure resource_settings points to the correct subclass of ResourceSettings.')


class ResourceNotFound(StandardError):
    status_code = 422



import pymongo

from ereuse_devicehub.resources.account.role import Role
from ereuse_devicehub.resources.resource import ResourceSettings
from ereuse_devicehub.resources.schema import Thing
from ereuse_devicehub.validation.coercer import Coercer
from ereuse_devicehub.validation.validation import ALLOWED_WRITE_ROLES



class Event(Thing):
    date = {
        'type': 'datetime',  # User specified date when the event was triggered
        'sink': -2,
        'description': 'When this happened. Leave blank if it is happening now'
    }
    secured = {
        'type': 'boolean',
        'default': False,
        'sink': -3
    }
    incidence = {
        'type': 'boolean',
        'default': False,
        'sink': -3,
        'description': 'Check if something went wrong, you can add details in a comment'
    }
    comment = {
        'type': 'string',
        'sink': -4,
        'description': 'Short comment for fast and easy reading'
    }
    byUser = {
        'type': 'objectid',
        'data_relation': {
            'resource': 'accounts',
            'field': '_id',
            'embeddable': True
        },
        'sink': 2,
        'coerce_with_context': Coercer.url_to_id,
        ALLOWED_WRITE_ROLES: Role.MANAGERS
    }
    byOrganization = {  # Materialization of the organization that, by the time of the event, the user worked in
        'type': 'string',
        'readonly': True
    }


class EventSettings(ResourceSettings):
    resource_methods = ['GET']
    item_methods = ['GET', 'DELETE']
    _schema = Event  # We update the schema in DOMAIN
    datasource = {
        'source': 'events',
        'default_sort': [('_created', -1)]
    }
    mongo_indexes = {
        '@type': [('@type', pymongo.DESCENDING)],
        'device': [('device', pymongo.HASHED)],
        'components': [('components', pymongo.DESCENDING)],
    }
    cache_control = 'max-age=15, must-revalidate'
    glyphicon = 'camera'



import datetime
import time
from datetime import timedelta
from functools import update_wrapper
from functools import wraps
from wsgiref.handlers import format_date_time

from flask import make_response, request, current_app
from requests.compat import basestring


def cache(expires=None, round_to_minute=False):
    """
        Add Flask cache response headers based on expires in seconds.

        If expires is None, caching will be disabled.
        Otherwise, caching headers are set to expire in now + expires seconds
        If round_to_minute is True, then it will always expire at the start of a minute (seconds = 0)

        Example usage:

        @app.route('/map')
        @cache(expires=60)
        def index():
          return render_template('index.html')
    """

    def cache_decorator(view):
        @wraps(view)
        def cache_func(*args, **kwargs):
            now = datetime.datetime.now()

            response = make_response(view(*args, **kwargs))
            response.headers['Last-Modified'] = format_date_time(time.mktime(now.timetuple()))

            if expires is None:
                response.headers[
                    'Cache-Control'] = 'no-store, no-cache, must-revalidate, post-check=0, pre-check=0, max-age=0'
                response.headers['Expires'] = '-1'
            else:
                expires_time = now + datetime.timedelta(seconds=expires)

                if round_to_minute:
                    expires_time = expires_time.replace(second=0, microsecond=0)

                response.headers['Cache-Control'] = 'public'
                response.headers['Expires'] = format_date_time(time.mktime(expires_time.timetuple()))

            return response

        return cache_func

    return cache_decorator


def crossdomain(origin=None, methods=None, headers=None,
                max_age=21600, attach_to_all=True,
                automatic_options=True):
    if methods is not None:
        methods = ', '.join(sorted(x.upper() for x in methods))
    if headers is not None and not isinstance(headers, basestring):
        headers = ', '.join(x.upper() for x in headers)
    if not isinstance(origin, basestring):
        origin = ', '.join(origin)
    if isinstance(max_age, timedelta):
        max_age = max_age.total_seconds()

    def get_methods():
        if methods is not None:
            return methods

        options_resp = current_app.make_default_options_response()
        return options_resp.headers['allow']

    def decorator(f):
        def wrapped_function(*args, **kwargs):
            if automatic_options and request.method == 'OPTIONS':
                resp = current_app.make_default_options_response()
            else:
                resp = make_response(f(*args, **kwargs))
            if not attach_to_all and request.method != 'OPTIONS':
                return resp

            h = resp.headers

            h['Access-Control-Allow-Origin'] = origin
            h['Access-Control-Allow-Methods'] = get_methods()
            h['Access-Control-Max-Age'] = str(max_age)
            if headers is not None:
                h['Access-Control-Allow-Headers'] = headers
            return resp

        f.provide_automatic_options = False
        return update_wrapper(wrapped_function, f)

    return decorator



import datetime

from ereuse_devicehub.resources.account.domain import AccountDomain, UserNotFound
from ereuse_devicehub.resources.event.device import DeviceEventDomain
from ereuse_devicehub.rest import execute_post
from ereuse_devicehub.utils import Naming


def transfer_property(receives: list):
    for receive in receives:
        if receive['automaticallyAllocate']:
            allocate_type = DeviceEventDomain.new_type('Allocate')
            a = execute_post(Naming.resource(allocate_type), {
                '@type': allocate_type,
                'to': receive['receiver'],
                'devices': receive['devices']
            })
            receive['_created'] = receive['_updated'] = a['_created'] + datetime.timedelta(milliseconds=1)


def set_organization(receives: list):
    """
    This method needs to execute after add_or_get_inactive_account
    :param receives:
    :return:
    """
    for receive in receives:
        try:
            receive['receiverOrganization'] = AccountDomain.get_one(receive['receiver']).get('organization')
        except UserNotFound:
            pass



from datetime import datetime

from ereuse_devicehub.exceptions import InnerRequestError
from ereuse_devicehub.resources.event.device import DeviceEventDomain
from ereuse_devicehub.resources.event.domain import EventNotFound
from ereuse_devicehub.rest import execute_delete
from ereuse_devicehub.utils import Naming
from flask import current_app as app
from flask import request, g

from .snapshot import Snapshot


def on_insert_snapshot(items):
    for item in items:
        if 'label' in item:
            item['device']['labelId'] = item['label']  # todo as we do not update the values of a device,
        # todo we will never udpate, thus materializing new label ids
        snapshot = Snapshot(item['device'], item['components'])
        item['events'] = [new_events['_id'] for new_events in snapshot.execute()]
        item['device'] = item['device']['_id']
        item['components'] = [component['_id'] for component in item['components']]
        item['unsecured'] = snapshot.unsecured
        from ereuse_devicehub.resources.hooks import set_date
        set_date(None, items)  # Let's get the time AFTER creating the other events


def save_request(items):
    """
    Saves the original request in a string in the 'request' field in json for debugging purposes.

    Warning: This method does not support bulk inserts.
    """
    items[0]['request'] = request.data.decode()


def materialize_test_hard_drives(snapshots: list):
    for i, test_hard_drives in g.snapshot_test_hard_drives:
        _materialize_event_in_device(test_hard_drives, 'tests')


def materialize_erase_basic(snapshots: list):
    for i, erase_basic in g.snapshot_basic_erasures:
        _materialize_event_in_device(erase_basic, 'erasures')


def _materialize_event_in_device(event, field_name):
    app.data.driver.db.devices.update({'_id': event['device']}, {'$push': {field_name: event['_id']}})


def set_secured(snapshots: list):
    """
    Sets secured param in a snapshot if this was signed.
    :param snapshots:
    :return:
    """
    for snapshot in snapshots:
        snapshot['secured'] = g.trusted_json


def delete_events(resource_name: str, snapshot: dict):
    """
    Deletes the events that were created with the snapshot.
    """
    if snapshot.get('@type') == 'devices:Snapshot':
        for event_id in snapshot['events']:
            try:
                # If the first event is Register, erasing the device will erase the rest of events
                event = DeviceEventDomain.get_one(event_id)
            except EventNotFound:
                pass
            else:
                try:
                    execute_delete(Naming.resource(event['@type']), event['_id'])
                except InnerRequestError as e:
                    if e.status_code != 404:
                        raise e



from ereuse_devicehub.resources.event.device.domain import DeviceEventDomain
from ereuse_devicehub.resources.event.device.settings import EventWithDevices, EventSubSettingsMultipleDevices
from ereuse_devicehub.resources.resource import Resource

"""
    Registers the generic types for the events.

    This module creates as many Schema (extending from EventWithDevices) and Settings (extending from
    EventSubSettingsMultipleDevices) as generic types are.
"""

for (generic_type, settings) in DeviceEventDomain.GENERIC_TYPES.items():
    Resource.create(generic_type, EventWithDevices, {}, EventSubSettingsMultipleDevices, settings)



from assertpy import assert_that
from ereuse_devicehub.tests.test_resources.test_events.test_device_event import TestDeviceEvent


class TestAllocate(TestDeviceEvent):
    def test_create_allocate_with_place(self):
        allocate = self.get_fixture('allocate', 'allocate')
        to = self.get_first('accounts')['_id']
        allocate['to'] = to
        allocate['devices'] = self.devices_id
        allocate = self.post_and_check(self.DEVICE_EVENT + '/allocate', allocate)
        for device_id in self.devices_id:
            device, _ = self.get(self.DEVICES, '', device_id)
            assert_that(device['owners']).contains(to)
        return allocate

    def test_delete_allocate(self):
        allocate = self.test_create_allocate_with_place()
        self.delete_and_check(self.DEVICE_EVENT + '/allocate/' + allocate['_id'])
        _, status = self.get(self.EVENTS, '', allocate['_id'])
        self.assert404(status)
        for device_id in self.devices_id:
            device, _ = self.get(self.DEVICES, '', device_id)
            assert_that(device['owners']).is_empty()

    def test_delete_allocate_deallocate(self):
        # Now let's create again another allocate and let's add a Remove, then let's delete the remove
        allocate = self.test_create_allocate_with_place()
        data = {'@type': 'devices:Deallocate', 'from': allocate['to'], 'devices': self.devices_id}
        deallocate = self.post_and_check(self.DEVICE_EVENT + '/deallocate', data)
        for device_id in self.devices_id:
            device, _ = self.get(self.DEVICES, '', device_id)
            assert_that(device['owners']).is_empty()
        # Now let's remove the deallocate. We need to re-apply the materialization in owners
        self.delete_and_check(self.DEVICE_EVENT + '/deallocate/' + deallocate['_id'])
        for device_id in self.devices_id:
            device, _ = self.get(self.DEVICES, '', device_id)
            assert_that(device['owners']).contains(allocate['to'])

    def test_allocate_delete_complex(self):
        # Let's make the following:
        # Allocate account A, Allocate acc B, Deallocate A, Allocate again A
        allocate_a = self.test_create_allocate_with_place()
        allocate_b = self.get_fixture('allocate', 'allocate')
        allocate_b['devices'] = self.devices_id
        allocate_b['to'] = {'email': 'b@b.com'}
        allocate_b = self.post_and_check(self.DEVICE_EVENT + '/allocate', allocate_b)
        deallocate_a = {'@type': 'devices:Deallocate', 'from': allocate_a['to'], 'devices': self.devices_id}
        deallocate_a = self.post_and_check(self.DEVICE_EVENT + '/deallocate', deallocate_a)
        allocate_a_2 = self.test_create_allocate_with_place()
        # Let's erase the first Allocate A. The result in 'owners' should be A (for the second allocate) and B.
        self.delete_and_check(self.DEVICE_EVENT + '/allocate/' + allocate_a['_id'])
        for device_id in self.devices_id:
            device, _ = self.get(self.DEVICES, '', device_id)
            assert_that(device['owners']).contains(allocate_a_2['to'], allocate_b['to'])

        # We should have deleted the first Allocate A and deallocate B
        _, status = self.get(self.EVENTS, '', allocate_a['_id'])
        self.assert404(status)
        _, status = self.get(self.EVENTS, '', deallocate_a['_id'])
        self.assert404(status)
        _, status = self.get(self.EVENTS, '', allocate_b['_id'])
        self.assert200(status)
        _, status = self.get(self.EVENTS, '', allocate_a_2['_id'])
        self.assert200(status)



from assertpy import assert_that
from ereuse_devicehub.resources.device.schema import Device
from ereuse_devicehub.resources.device.settings import DeviceSubSettings
from ereuse_devicehub.resources.resource import Resource
from ereuse_devicehub.tests import TestStandard
from ereuse_devicehub.utils import Naming


class TestResource(TestStandard):
    dummy_device_name = 'DummyDevice'
    dummy_field = 'dummyField'
    dummy_device_class = {dummy_field: {'type': 'string'}}
    dummy_device_settings = {dummy_field: 'just to try'}

    def setUp(self, settings_file=None, url_converters=None):
        """
        Creates a dummy sub-type of device. Note that we need to do it before instantiating the app.
        :param settings_file:
        :param url_converters:
        :return:
        """
        self.create_dummy_type_of_device()
        super().setUp(settings_file, url_converters)

    def create_dummy_type_of_device(self):
        # todo this method will fail if TestResource is not the **first** test executed
        Resource.create(self.dummy_device_name, Device, self.dummy_device_class, DeviceSubSettings,
                        self.dummy_device_settings)

    def test_resource(self):
        """
        Creates a resource endpoint extending ResourceSettings and RDFS, checking that both processes are done correctly,
        performing requests to the endpoint.
        """
        # The endpoint is defined in setUp and created as part of the regular workflow in DeviceHub.
        # Let's validate it.
        resource_name = Naming.resource(self.dummy_device_name)
        self.assertIn(resource_name, self.domain)
        settings = self.domain[resource_name]
        assert_that(self.dummy_device_class).is_subset_of(settings['schema'])
        assert_that(self.dummy_device_settings).is_subset_of(settings)
        # Let's check we can perform some actions
        # Let's create a dummy device
        dummy_device = {'serialNumber': '33', 'model': 'model1', 'manufacturer': '234', self.dummy_field: 'dummyField',
                        '@type': self.dummy_device_name}
        register = {'@type': 'Register', 'device': dummy_device}
        self.post_and_check('{}/{}'.format(self.DEVICE_EVENT, 'register'), register)
        # And now without the necessary new field
        # wrong_dummy_device = dummy_device
        # del wrong_dummy_device[self.dummy_field]
        # _, status_code = self.post(self.DEVICES, wrong_dummy_device)
        # self.assert422(status_code)
        # Let's get a collection of dummy devices
        _, status_code = self.get(self.DEVICES)
        self.assert200(status_code)
        # Let's get the first dummy device
        dummy_device = self.get_first(self.DEVICES)
        dummy_device['@type'] = self.dummy_device_name
        # Let's try posting other devices, so we know our new resource doesn't affect others
        self.get_fixtures_computers()



import copy
import inspect
from collections import Sequence

from ereuse_devicehub.utils import Naming


class Resource:
    """
        Base class for Resources.

        Resources are the URL endpoints of the API, and represent the ontology of DeviceHub. Python-EVE expects us to
        deliver a dictionary with settings of how to configure the resource, including the schema definition.

        We have encapsulated resource settings and their schemas into classes that extend this one, Resource. This is
        our way to implement inheritance between resource schemas and resource settings. An example illustrates better
        the idea:
                                Resource
                    /                       \
                  RDFS                ResourceSettings
                    |           /            |              \
                  Thing       DeviceSettings EventSettings ...
                /   |    \
            Device Event ...

        Thanks to this we do not need to repeat all over again inherited properties.

        Resource provides methods to perform the inheritance, as it is not a classic (pythonic for example) 100%
        inheritance. Please, refer to the methods to get more information.

        Resource is extended by :class:`app.resources.resource.ResourceSettings`
        (which takes care of the settings provided to python-EVE) and
        :class:`app.resources.schema.RDFS` (which provides used RDFS properties, and takes care of the schemas).

        To get a python-EVE ready dictionary, just instantiate the class.
    """

    @staticmethod
    def __new__(cls) -> dict:
        """
            Gets a dictionary with the fields of the resource, ready to use in Python-EVE.
        """
        raise NotImplementedError()

    @classmethod
    def superclasses_fields(cls, top_classes_to_ignore: int) -> dict:
        """
            Returns the superclasses attributes and the actual one.

            This method replaces the attributes of the superclasses that are found in the subclasses.
            This means, for example, that attributes in the actual class with have preference over the one of the
            superclass.
            :param top_classes_to_ignore: Set the top classes to exclude from the attribute gathering. For example, if
            we do not want the values from Object class, as it is the top class, we would set it to 1.
        """
        fields = {}
        for superclass in reversed(cls.superclasses(top_classes_to_ignore)):
            fields.update(superclass.actual_fields())
        return fields

    @classmethod
    def _clean(cls, attributes: dict, attributes_to_remove: Sequence = None) -> dict:
        """
            Obtains a fully new dict only the interesting attributes from a resource.
            This is, removing built-in attributes from python, callables (methods...) and specified attributes.

            Extend this method to do other clean-up stuff with the attributes.
            :param attributes_to_remove: A Collection of attribute names to be removed.
            :returns: A deep copied dictionary
        """
        attributes_to_remove = [] if attributes_to_remove is None else attributes_to_remove
        for key in dict(attributes).keys():
            if key.startswith('__') or key in attributes_to_remove:
                del attributes[key]
            else:
                attribute = getattr(cls, key)
                if callable(attribute) and not (inspect.isclass(attribute) and issubclass(attribute, Resource)):
                    del attributes[key]

        return copy.deepcopy(attributes)

    @classmethod
    def subclasses(cls) -> list:
        """Obtains the subclasses of the actual class, without the actual class."""
        subclasses = []
        for subclass in cls.__subclasses__():
            subclasses.append(subclass)
            subclasses.extend(subclass.subclasses())
        return subclasses

    @classmethod
    def superclasses(cls, ignore_n: int) -> tuple:
        """
        :param ignore_n: The number of top superclasses to ignore.
        """
        return inspect.getmro(cls)[:-ignore_n]

    @classmethod
    def actual_fields(cls):
        return cls._clean(dict(vars(cls)), ())

    @classmethod
    def create(cls, name: str, parent_schema: object, schema: dict, parent_resource_settings: object,
               resource_settings: dict):
        """
        Defines a new resource, setting its endpoint settings and schema.

        Use this method before instantiating the app (app = DeviceHub()). Future work is to avoid this restriction.
        :param name: The name of the resource to create. We append 'Settings' for the settings of the resource.
        :param parent_schema: The parent class to extend, minimum RDFS.
        :param schema: A dict personalizing the schema, or empty dict if using exactly the same schema of the parent.
        :param parent_resource_settings: As 'parent_schema', the ResourceSettings class or subclass to extend from.
        :param resource_settings: As 'schema', for the resource settings.
        """
        # Although it is not very pythonic to register in globals, we are doing so at initialization
        globals()[name] = type(name, (parent_schema,), schema)
        resource_settings_name = '{}Settings'.format(name)
        resource_settings['_schema'] = globals()[name]
        globals()[resource_settings_name] = type(resource_settings_name, (parent_resource_settings,), resource_settings)


class ResourceSettings(Resource):
    # Custom defaults (apart from defined defaults in EVE's settings)
    use_default_database = False
    """Use the user's specific databases or the common default one"""
    extra_response_fields = ['@type', 'label', 'url', 'sameAs', 'description']
    _schema = False
    cache_control = 'max-age=3, must-revalidate'

    @staticmethod
    def __new__(cls) -> dict:
        fields = cls.superclasses_fields(2)  # We ignore Object, Resource
        # We make possible GET the resource (not item) /events/devices/snapshot by providing
        # a default filter. Note that first-level resources (devices) do not use it as there is no need
        # todo write tests for this
        if 'datasource' in fields and len(cls.superclasses(2)) > 2:
            if len(cls.sub_resources()) == 0:
                fields['datasource']['filter'] = {'@type': cls._schema.type_name}
            else:
                fields['datasource']['filter'] = {'@type': {'$in': list(cls._schema.types)}}
        if fields['_schema']:  # We get the schema. This is a costly operation we do not want to do in actual_fields
            fields['schema'] = fields['_schema']()
        return fields

    @classmethod
    def actual_fields(cls):
        """
            Gets only the actual fields of the class (ignoring parents'), alongside with the URL.
        """
        if not hasattr(cls, '_schema'):
            raise TypeError('Resource does not have any schema')
        fields = super(ResourceSettings, cls).actual_fields()
        super_resources = super(ResourceSettings, cls).superclasses(2)
        # Creating the url for the resource
        names = []
        for resource in reversed(super_resources):
            if getattr(resource, '_schema', False):
                # We get the URL: first by getting settings['url'] or using the name of the class
                names.append(resource._schema._settings.get('url', Naming.resource(resource._schema.__name__)))
        fields['url'] = '/'.join(names)
        return fields

    @classmethod
    def sub_resources(cls):
        """
        Returns all the sub-resources, without including the actual resource
        """
        return [subclass for subclass in cls.subclasses() if getattr(subclass, '_schema', False)]

    @classmethod
    def resource_name(cls):
        return cls._schema.resource_name



from ereuse_devicehub.resources.event.device.settings import parent, EventWithOneDevice, EventSubSettingsOneDevice
from .step_settings import step


class EraseBasic(EventWithOneDevice):
    secureRandomSteps = {
        'type': 'natural',
        'required': True
    }
    endingTime = {
        'type': 'datetime'
    }
    startingTime = {
        'type': 'datetime'
    }
    success = {
        'type': 'boolean',
    }
    cleanWithZeros = {
        'type': 'boolean'
    }
    steps = {
        'type': 'list',  # OrderedSet
        'schema': {
            'type': 'dict',
            'schema': step
        }
    }
    parent = parent


class EraseBasicSettings(EventSubSettingsOneDevice):
    _schema = EraseBasic
    fa = 'fa-eraser'
    short_description = 'Fast erasure of the HardDrive'
    item_methods = ['GET']
